package ims.data;

import java.util.Iterator;
import java.util.List;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.oracle.MinimalConfiguration;
import ims.trans.transitions.ITransition;

public class Instance {
	
	public static final int[] EMPTY_INT_ARR=new int[0];
	public static final int[] EMPTY_MORPH_ARR=new int[]{AbstractSymbolMapping.NONE_IDX};
	
	public static final int STAG_LEFT=0;
	public static final int STAG_RIGHT=1;
	
	public static final int STAG_TRUE=1;
	public static final int STAG_FALSE=0;
	
	public static final int[] STAG_DEP_NONE=new int[2];
	public static final int[] STAG_DEP_BOTH=new int[2];
	public static final int[] STAG_DEP_LEFT=new int[2];
	public static final int[] STAG_DEP_RIGHT=new int[2];
	
	static { //Assign these arrays
		STAG_DEP_NONE[0]=STAG_FALSE;STAG_DEP_NONE[1]=STAG_FALSE;
		STAG_DEP_BOTH[0]=STAG_TRUE;STAG_DEP_BOTH[1]=STAG_TRUE;
		STAG_DEP_LEFT[STAG_LEFT]=STAG_TRUE; STAG_DEP_LEFT[STAG_RIGHT]=STAG_FALSE;
		STAG_DEP_RIGHT[STAG_LEFT]=STAG_FALSE; STAG_DEP_RIGHT[STAG_RIGHT]=STAG_TRUE;
	}
	
	//Features at our disposal
	public final int[] form;
	public final int[] lemma;
	public final int[] tag;
	public final int[] ffeat;
	public final int[][] feat;
	
	public final int[] stackHead;
	public final int[] stackLbl;
	public final boolean[] stackHasLeftDep;
	public final boolean[] stackHasRightDep;

	public final int[] stag;
	public final int[] stagLbl;
	public final int[] stagHeadDir;
	public final int[][] stagDeps;
	
	public final int[] shape;
	public final int[] firstCharShape;
	public final int[] lastCharShape;
	
	//What we're trying to predict (sort of)
	public final int[] head;
	public final int[] lbl;
	
	//Trans sequence
	public final int[] transSequence; //Canonical;
	public int[] lastTRSeq;           //Last one
	public int[] fileTrSeqMapped;           //When using a file
	public String[] fileTrSeq;
	
	public final int sentId;
	public final TIntSet sentenceBegin=new TIntHashSet();
	
	public Instance(int[] heads){
		this(heads.length,null,-1,false,false,false);
	}
	
	public Instance(int length,int[] transSequence,int sentId,boolean stag,boolean stack,boolean shape){ //length includes root token
		this.sentId=sentId;
		form=new int[length];
		lemma=new int[length];
		tag=new int[length];
		ffeat=new int[length];
		feat=new int[length][];
		
		head=new int[length];
		lbl=new int[length];
		this.transSequence=transSequence;
		if(stag){
			this.stag=new int[length];
			this.stagLbl=new int[length];
			this.stagHeadDir=new int[length];
			this.stagDeps=new int[length][];
		}else{
			this.stag=null;
			this.stagLbl=null;
			this.stagHeadDir=null;
			this.stagDeps=null;
		}
		if(stack){
			this.stackHead=new int[length];
			this.stackLbl=new int[length];
			this.stackHasLeftDep=new boolean[length];
			this.stackHasRightDep=new boolean[length];
		} else {
			this.stackHead=null;
			this.stackLbl=null;
			this.stackHasLeftDep=null;
			this.stackHasRightDep=null;
		}
//		if(shape){
			this.shape=new int[length];
			this.firstCharShape=new int[length];
			this.lastCharShape=new int[length];
//		} else {
//			this.shape=null;
//			this.firstCharShape=null;
//			this.lastCharShape=null;			
//		}
	}


	public static Instance mergeSentInstancesToOne(List<Instance> sentInsts, boolean useStag, boolean useStack,boolean useShape, int sentId,boolean blockingSB, SymbolTable symTab, boolean training) {
		if(useStag || useStack){
			throw new Error("TODO: implement this stuff for stagging and stacking");
		}
		//Need to create the full seq and count the length to begin with
		int totLength=1; //1 for root
		for(Instance inst:sentInsts){
			totLength+=inst.form.length-1;
		}
		
		int[] trSeq;
		//If training
		if(training){
			trSeq=mergeTRSeqForDocTraining(sentInsts,blockingSB,symTab);
		} else {
			trSeq=null;
		}

		//Ok, now we can actually create the new instance and start copying stuff over
		Instance inst=new Instance(totLength, trSeq, sentId, useStack, useStack, useShape);
		int tokOffset=0;
		int dest=1;
		for(Instance i:sentInsts){
			inst.sentenceBegin.add(dest);
			if(tokOffset==0){ //Copy root from the first sentence
				inst.form[0]=i.form[0];
				inst.lemma[0]=i.lemma[0];
				inst.tag[0]=i.tag[0];
				inst.feat[0]=i.feat[0];
				inst.ffeat[0]=i.ffeat[0];
				inst.head[0]=i.head[0];
				inst.lbl[0]=i.lbl[0];
			}
			System.arraycopy(i.form, 1, inst.form, dest, i.form.length-1);
			System.arraycopy(i.lemma, 1, inst.lemma, dest, i.lemma.length-1);
			System.arraycopy(i.tag, 1, inst.tag, dest, i.tag.length-1);
			System.arraycopy(i.ffeat, 1, inst.ffeat, dest, i.ffeat.length-1);
			System.arraycopy(i.feat, 1, inst.feat, dest, i.feat.length-1);
			System.arraycopy(i.lbl, 1, inst.lbl, dest, i.lbl.length-1);
			for(int q=1;q<i.head.length;++q){
				if(i.head[q]==0)
					inst.head[dest+q-1]=0;
				else
					inst.head[dest+q-1]=i.head[q]+tokOffset;
			}
			tokOffset+=i.form.length-1;
			dest+=i.form.length-1;
		}
		return inst;
	}


	private static int[] mergeTRSeqForDocTraining(List<Instance> sentInsts,	boolean blockingSB, SymbolTable symTab) {
		
		//count length of new trseq
		int totTRs=sentInsts.size()-1; //n-1 SB transitions required
		for(Instance inst:sentInsts){
			totTRs+=inst.transSequence.length;
		}

		//Now build the new TR seq
		int trOffset=0;
		int[] trSeq=new int[totTRs];
		int sbTRid=symTab.transitions.lookup(TR.SB.toString());
		if(blockingSB){
			int shTRid=symTab.transitions.lookup(TR.SH.toString());
			//Here we copy the individual transition sequence up until the last shift, and then insert the BSB here
			for(Iterator<Instance> it=sentInsts.iterator();it.hasNext();){
				Instance inst=it.next();
				
				int lastShift=0;
				
//				//XXX This was the old buggy implementation, where we would go on until the last Shift
//				//    Then we would miss the case when we reach an empty buffer and then swap stuff back (fixed below)
//				for(int k=0;k<inst.transSequence.length;++k){
//					if(inst.transSequence[k]==shTRid)
//						lastShift=k;
//				}

//				//XXX this code fixes the swap issue
				MinimalConfiguration mc=new MinimalConfiguration(inst.head);
				ITransition[] index=symTab.getITransionIndex();
				for(int k=0;k<inst.transSequence.length;++k){
					TR t=index[inst.transSequence[k]].getTR();
					mc.applyTR(t);
					if(mc.bufferSize()==0){
						lastShift=k;
						break;
					}
				}
				
				System.arraycopy(inst.transSequence, 0, trSeq, trOffset, lastShift+1); //Copy until and including the last shift
				trOffset+=lastShift+1;   //accounts for the copied transitions
				if(it.hasNext()){ //only add the BSB if there is another sentence following
					trSeq[trOffset]=sbTRid; //insert the BSB
					++trOffset;              //increment trOffset to account for the BSB
				}
				System.arraycopy(inst.transSequence, lastShift+1, trSeq, trOffset, inst.transSequence.length-lastShift-1); //Copy remaining reduce operations
				trOffset+=inst.transSequence.length-lastShift-1; //increment trOffset to cover remaining reduce
			}
		} else {
			for(Instance inst:sentInsts){
				System.arraycopy(inst.transSequence, 0, trSeq, trOffset, inst.transSequence.length);
				trOffset+=inst.transSequence.length;
				if(trOffset!=trSeq.length){ //For the last sentence we don't add this transition
					trSeq[trOffset]=sbTRid;
					trOffset++;
				}
			}
		}
		if(trOffset!=trSeq.length)
			throw new Error("this shouldn't happen if I did the math right");

		return trSeq;
	}
	
}
