package ims.hotcoref.symbols;

import ims.util.DBO;

import java.io.Serializable;

public class TagSymbolTable implements Serializable {

	private static final long serialVersionUID = -7278617304486679104L;
	
	public StringSymbolMapping 
		forms,
		affixes,
		tags;
	
	
	private boolean initialized=false;
	private boolean frozen=false;
	
	public int[] tagIdx;
	private String[] tagLookup;

	public TagSymbolTable(){
		initTable();
	}

	private void initTable() {
		if(initialized)
			throw new UnsupportedOperationException("cant init twice");
		initialized=true;
		forms=new StringSymbolMapping("Form");
		affixes=new StringSymbolMapping("Affixes");
		tags=new StringSymbolMapping("Tags");
	}
	
	public void freezeMappings(){
		if(frozen)
			throw new UnsupportedOperationException("cant freeze twice");
		DBO.println("Freezing symbol table");
		DBO.printlnNoPrefix("Types");
		
		if(this.forms!=null) 		{DBO.printlnNoPrefix(String.format("Forms:         %5d : %2d", forms.getItems(),forms.getBits())); 				forms.freeze();             }
		if(this.affixes!=null) 		{DBO.printlnNoPrefix(String.format("Affixes:       %5d : %2d", affixes.getItems(),affixes.getBits()));			    affixes.freeze();			}
		if(this.tags!=null) 		{DBO.printlnNoPrefix(String.format("Tags:          %5d : %2d", tags.getItems(),tags.getBits()));				tags.freeze();				}
		
		tagIdx=new int[tags.getItems()-AbstractSymbolMapping.INIT_IDX];
		tagLookup=tags.getInverted();
		for(int t=AbstractSymbolMapping.INIT_IDX,i=0;i<tagIdx.length;++i,++t){
			tagIdx[i]=t;
		}
		
		frozen=true;
		DBO.println();
	}
	
	public String lookupTagString(int t){
		return tagLookup[t];
	}
}
