package ims.io;

import ims.data.ISentence;

import java.util.Iterator;

public class ReversingReader<T extends ISentence> implements SentenceReader<T> {

	private final SentenceReader<T> reader;
	
	public ReversingReader(SentenceReader<T> reader){
		this.reader=reader;
	}
	
	public static <T extends ISentence> ReversingReader<T> createReader(SentenceReader<T> reader){
		return new ReversingReader<T>(reader);
	}
	
	@Override
	public Iterator<T> iterator() {
		return new SentenceIterator<T>(reader.iterator());
	}
	
	static class SentenceIterator<T extends ISentence> implements Iterator<T> {

		private final Iterator<T> it;
		
		SentenceIterator(Iterator<T> it){
			this.it=it;
		}
		
		@Override
		public boolean hasNext() {
			return it.hasNext();
		}

		@Override
		public T next() {
			T s=it.next();
			s.reverseSentence();
			return s;
		}
		
		@Override
		public final void remove() {
			throw new UnsupportedOperationException("Remove not supported");
		}
		
	}

}
