package ims.oracle;

import ims.data.Buffer;
import ims.data.Stack;
import ims.trans.State;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

public class Configuration implements Comparable<Configuration>,IConfiguration{

	public static enum TR {
		
		SH('a',"Shift","#CCCC66"),      //Yellowish 
		LA('b',"LeftArc","#6666CC"),    //Bluish
		RAS('c',"RightArcS","#66CC66"), //Greenish
//		RAE('d',"RightArcE","#66CC88"), //Greenish
//		RE('e',"Reduce","#66CCCC"),     //Cyanish
		SW('f',"Swap","#CC6666"),       //Reddish
//		SB('g',"Sent","#000000"),
		SB('h',"SentBound","#000000");
		
		public final char k;
		public final String name;
		public final String htmlColor;
		TR(char k,String name,String htmlColor){
			this.k=k;
			this.name=name;
			this.htmlColor=htmlColor;
		}
		
		public static TR fromC(char c){
			for(TR tr:TR.values())
				if(tr.k==c)
					return tr;
			return null;
		}
		
		public static TR fromString(String s){
			if(s.contains("_")){
				String[] a=s.split("_",2);
				return TR.valueOf(a[0]);
			} else {
				return TR.valueOf(s);
			}
		}
	}
	
	public int x=-1;
	public int y=-1;
	
	private Stack stack;
	private Buffer buffer;
	
	private int[] predHeads;
	
	private short[] remainingDepsLeft;
	private short[] remainingDepsRight;
	
	private String signature="0";
	
	private boolean[] onStack; //Fast lookup if a token is on buffer or stack (or gone)
	
	
	int totalTransitions=0;
	
	public short deadEndState=0;
	
	public Map<Configuration,TR> incoming=new HashMap<Configuration,TR>();
	public Map<TR,Configuration> outgoing=new HashMap<TR,Configuration>();
	
	
	public Configuration(int[] heads){
		stack=new Stack();
		buffer=new Buffer(heads.length-1);
		onStack=new boolean[heads.length];
		onStack[0]=true;
		predHeads=new int[heads.length];
		Arrays.fill(predHeads,-1);
		remainingDepsLeft=new short[heads.length];
		remainingDepsRight=new short[heads.length];
		for(int i=1;i<heads.length;++i){
			if(i<heads[i]) //head on the right, means it's a left dep of head
				remainingDepsLeft[heads[i]]++;
			else //head on the left, means it's a right dep of the head
				remainingDepsRight[heads[i]]++;
		}
//		swappedTokens=new boolean[heads.length];
	}
		
	private Configuration(){};
	
	/*
	 * We make a shallow copy here
	 */
	public Configuration copy(){
		Configuration r=new Configuration();
//		r.stack=stack.copy();
//		r.buffer=buffer.copy();
//		r.treeSig=treeSig.copy();
//		r.predHeads=Arrays.copyOf(predHeads, predHeads.length);
//		r.swappedTokens=Arrays.copyOf(swappedTokens, swappedTokens.length);
//		r.remainingDepsLeft=Arrays.copyOf(remainingDepsLeft,remainingDepsLeft.length); //Also with these guys we can do lazy copying, but we should also fix rightArc and leftArc to check that we didn't swap past what we wanted (so a left dependent is attached through a rightarc)
//		r.remainingDepsRight=Arrays.copyOf(remainingDepsRight,remainingDepsRight.length);
		r.stack=stack;
		r.buffer=buffer;
//		r.treeSig=treeSig;
		r.predHeads=predHeads;
		r.signature=signature;
//		r.swappedTokens=swappedTokens;
		r.remainingDepsLeft=remainingDepsLeft;
		r.remainingDepsRight=remainingDepsRight;
		r.signature=signature;
		r.onStack=onStack;
		return r;
	}
	
	public boolean equals(Object o){
		return equals((Configuration) o);
	}
	
	public boolean equals(Configuration o){
		if(o.stack.equals(stack) && 
		   o.buffer.equals(buffer) && 
		   Arrays.equals(o.predHeads, predHeads)
//		   && Arrays.equals(swappedTokens, this.swappedTokens) //This is bad!
		   )
			return true;
		else
			return false;
	}
	
	
	public int hashCode(){
		int h=buffer.get(0)*13+stack.size()*7;
		return h;
	}

	public void doShift() {
		buffer=buffer.copy();
		stack=stack.copy();
		onStack=Arrays.copyOf(onStack, onStack.length);
		int b0=buffer.removeFront();
		stack.add(b0);
		onStack[b0]=true;
		signature+=TR.SH.k;
		++totalTransitions;
	}

	public void doRightArcS() {
		predHeads=Arrays.copyOf(predHeads, predHeads.length);
		remainingDepsRight=Arrays.copyOf(remainingDepsRight, remainingDepsRight.length);
		stack=stack.copy();
		int dep=stack.pop();
		int head=stack.peek(0);
		remainingDepsRight[head]--;
		signature+=TR.RAS.k;
		predHeads[dep]=head;
		++totalTransitions;
	}

	public void doLeftArc() {
		predHeads=Arrays.copyOf(predHeads, predHeads.length);
		remainingDepsLeft=Arrays.copyOf(remainingDepsLeft, remainingDepsLeft.length);
		stack=stack.copy();
		int head=stack.pop();
		int dep=stack.pop();
		stack.add(head);
		remainingDepsLeft[head]--;
		signature+=TR.LA.k;
		predHeads[dep]=head;
		++totalTransitions;
	}	

	public void doSwap() {
		buffer=buffer.copy();
		stack=stack.copy();
		onStack=Arrays.copyOf(onStack, onStack.length);
		int s0=stack.pop();
		int s1=stack.pop();
		stack.add(s0);
		buffer.pushBack(s1);
		onStack[s1]=false;
		signature+=TR.SW.k;
		++totalTransitions;
	}
	
	public boolean hasAllDeps(int idx){
		return remainingDepsLeft[idx]==0 && remainingDepsRight[idx]==0;
	}
	
	public boolean hasAllLeftDeps(int idx){
		return remainingDepsLeft[idx]==0;
	}
	
	public boolean hasAllRightDeps(int idx){
		return remainingDepsRight[idx]==0;
	}
	
	@Override
	public int compareTo(Configuration o) {
		int B=buffer.compareTo(o.buffer);
		if(B!=0)
			return B;
		
		int S=stack.compareTo(o.stack);
		if(S!=0)
			return S;
		
		//That's not enough, the predHeads need to be different too.
		for(int i=1;i<predHeads.length;++i)
			if(predHeads[i]!=o.predHeads[i])
				return predHeads[i]-o.predHeads[i];

		//should probably compared the swappedTokens array too, to be on the safe side.
		
		return 0;
			
	}

	public String toString(){
		StringBuilder sb=new StringBuilder();
		sb.append("Buffer: ").append(buffer.toString()).append('\n');
		sb.append("Stack:  ").append(stack.toString()).append('\n');
		sb.append("Sig: ").append(signature).append('\n');
		sb.append("State sig: ").append(getStateSignature());
		return sb.toString();
	}

	public void applyTR(TR tr) {
		switch(tr){
		case SH:  doShift(); break;
		case LA:  doLeftArc(); break;
		case RAS: doRightArcS(); break;
		case SW:  doSwap(); break;
		default: throw new Error("Unknown TR: "+tr);
		}
	}

	public boolean isTerminal(){
		return buffer.isEmpty() && stack.size()==1;
	}

	public String getTransitionSeqSignature() {
		return signature;
	}
	
	/*
	 * Previously I had stack + SEP + buffer + SEP + tree. And SEP was '|' (which is ASCII 124). With sentences of length > 124 then the signatures could come out identical, because stack and buffer were appended after each other. I think by just changing the order it *should* be safe. But to be really safe let's also use a separator of much greater value 
	 */
	public String getStateSignature(){
		StringBuilder sb=new StringBuilder();
		buffer.appendSignature(sb).append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=1;i<predHeads.length;++i)
			sb.append((char) predHeads[i]);
//		treeSig.appendHeadSignature(sb).append(STATE_SIG_SEPARATOR);
		stack.appendSignature(sb);
//		for(int i=1;i<swappedTokens.length;++i) //This *SHOULD NOT* be part of the identity of a state
//			sb.append(swappedTokens[i]?'t':'f');
		return sb.toString();
	}

	@Override
	public int bufferSize() {
		return buffer.size();
	}

	@Override
	public boolean hasHead(int dep) {
		return predHeads[dep]!=-1;
	}

	@Override
	public int peekStack(int d) {
		return stack.peek(d);
	}

	@Override
	public int peekBuffer(int d) {
		return buffer.get(d);
	}

	@Override
	public boolean hasMinTwoStackElements() {
		return stack.size()>=2;
	}

	public int[] getPredHeads() {
		return predHeads;
	}

	public boolean isPermissible(TR tr) {
		switch(tr){
		case SH: return buffer.size()>0;
		case LA: return stack.size()>1 && predHeads[stack.peek(0)]==-1 && stack.peek(1)!=0;
		case SW: return stack.size()>1 && stack.peek(1)<stack.peek(0);
		case RAS: return stack.size()>1 && predHeads[stack.peek(0)]==-1;
		}
		return false;
	}

	public int stackSize() {
		return stack.size();
	}

	public String stackToString() {
		return stack.toString();
	}
	
	public String bufferToString() {
		return buffer.toString();
	}

	public String bufferToString(String[] t) {
		return buffer.toString(t);
	}

	@Override
	public TokenState getTokenState(int i) {
		if(!onStack[i])
			return TokenState.OnBuffer;
		if(hasHead(i))
			return TokenState.Gone;
		else
			return TokenState.OnStack;
	}

	@Override
	public int sentenceSize() {
		return predHeads.length;
	}

	@Override
	public int getBufferNoPushedBack() {
		return buffer.getBufferNoPushedBack();
	}

	@Override
	public int getHead(int d) {
		return predHeads[d];
	}

	@Override
	public int getLbl(int i) {
		throw new UnsupportedOperationException("not implemented");
	}

	@Override
	public StringBuilder appendStackSig(StringBuilder sb) {
		return appendStackSigLinear(sb,this);
	}

	@Override
	public String getULBLStateSig() {
		return State.getULBLSig(this);
	}

	@Override
	public String getLBLStateSig() {
		throw new UnsupportedOperationException("not implemented");
	}

	public static StringBuilder appendStackSigLinear(StringBuilder sb,IConfiguration c){
		for(int i=c.stackSize()-1;i>=0;--i)
			sb.append((char) c.peekStack(i));
		return sb;
	}

	@Override
	public int getRemainingDeps(int head) {
		return remainingDepsLeft[head]+remainingDepsRight[head];
	}

	@Override
	public Stack getStackCopy() {
		return stack.copy();
	}

	@Override
	public int getTotalTransitions() {
		return totalTransitions;
	}

	@Override
	public boolean isBlockedForShift() {
		throw new Error("not implemented");
	}

	@Override
	public TR getPrevTR() {
		throw new Error("not implemented");
	}

	@Override
	public boolean sbPermissible() {
		throw new Error("not implemented");
	}

}
