package ims.oracle;

import ims.data.Instance;
import ims.oracle.Configuration.TR;
import ims.trans.Options.TransitionSystem;

import java.util.ArrayList;
import java.util.List;

public class DeterministicArcStandardOracle extends AbstractStaticOracle {

	protected DeterministicArcStandardOracle(TR[] myTRs) {
		super(myTRs);
	}

	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels,boolean labeled) {
		return staticGetCanonicalSequence(heads,labels,labeled);
	}
	
	public static List<String> staticGetCanonicalSequence(int[] heads,String[] labels,boolean withLabels){
		Configuration c=new Configuration(heads);
		List<String> operations=new ArrayList<String>();
		while(!(c.bufferSize()==0 && c.stackSize()==1)){
			final String op;
			if(canLeftArc(heads,c)){
				op=(withLabels?TR.LA+"_"+labels[c.peekStack(1)]:TR.LA.toString());
				c.doLeftArc();
			} else if(canRightArc(heads,c)){
				op=(withLabels?TR.RAS+"_"+labels[c.peekStack(0)]:TR.RAS.toString());
				c.doRightArcS();
			} else {
				if(c.bufferSize()==0)
					return operations;
				op=TR.SH.toString();
				c.doShift();
			}
//			System.out.println(op);
			operations.add(op);
		}
		return operations;
	}
	
	public static boolean canRightArc(int[] heads,IConfiguration c){
		if(!c.hasMinTwoStackElements())
			return false;
		int stackTop=c.peekStack(0);
		int stackSecond=c.peekStack(1);
		if(heads[stackTop]==stackSecond && c.hasAllDeps(stackTop))
			return true;
		else
			return false;
	}
	
	public static boolean canLeftArc(int[] heads,IConfiguration c){
		if(!c.hasMinTwoStackElements())
			return false;
		int stackTop=c.peekStack(0);
		int stackSecond=c.peekStack(1);
		if(heads[stackSecond]==stackTop && c.hasAllDeps(stackSecond))
			return true;
		else
			return false;
	}

	@Override
	public boolean canDoNonProj() {
		return false;
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, Instance inst) {
		int[] heads=inst.head;
		switch(tr){
		case LA:  return canLeftArc(heads,conf);
		case RAS: return canRightArc(heads,conf);
		case SH:  return !canRightArc(heads,conf) && !canLeftArc(heads,conf);
		default:  throw new Error("not supported by this system");
		}
	}

	@Override
	public void initSentence(int[] heads) {
		
	}
	
	@Override
	public TransitionSystem getTS() {
		return TransitionSystem.ArcS;
	}
	
}
