package ims.oracle;

import ims.data.ISentence;
import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.GoldDynParser;
import ims.trans.Options.TransitionSystem;
import ims.trans.State;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFilterFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class DeterministicModelOracle extends AbstractStaticOracle {

	private final GoldDynParser gdp;
	private final int beam;
	
	protected DeterministicModelOracle(GoldDynParser gdp,int beam) {
		super(gdp.myTRs);
		this.gdp=gdp;
		this.beam=beam;
	}

	@Override
	public List<String> getCanonicalSequence(int[] heads,String[] labels,boolean labeled) {
		throw new UnsupportedOperationException("need sentence for this oracle");
	}

	@Override
	public List<String> getCanonicalSequence(ISentence sentence,boolean labeled) {
		if(labeled!=gdp.labeled)
			throw new Error("Model and parameters incompatible wrt to labels. Model: "+gdp.labeled+", Parameters: "+labeled);
		Instance inst=gdp.instanceMapping.createTestInstance(sentence, gdp.symTab);
		return getCanonicalSequence(inst);
	}
	
	private List<String> getCanonicalSequence(Instance inst) {
		State state=gdp.decode(inst,beam);
		List<String> r=state.recoverTRSeqStringList();
		return r;		
	}
	
	

	@Override
	public boolean canDoNonProj() {
		return Arrays.asList(myTRs).contains(TR.SW);
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, Instance inst) {
		int i=conf.getTotalTransitions(); //correct transition will be at this position in the cachedCanonicalSeq
		if(i>=cachedCanonicalSeq.size())
			return false;
		return cachedCanonicalSeq.get(i).equals(tr.toString());
	}

	private List<String> cachedCanonicalSeq;
	
	@Override
	public void initSentence(Instance inst){
		List<String> l=getCanonicalSequence(inst);
		this.cachedCanonicalSeq=new ArrayList<String>(l.size());
		for(String s:l){
			String t;
			if(s.contains("_")){
				t=s.split("_", 2)[0];
			} else {
				t=s;
			}
			this.cachedCanonicalSeq.add(t);
		}
	}

	@Override
	public void initSentence(int[] heads) {
		throw new UnsupportedOperationException("doesn't work with model oracle");
	}

	public ITransitionFilter getTransitionFilter(SymbolTable symTab, List<Instance> instances){
		return TransitionFilterFactory.getTransitionFilter(symTab, instances, gdp.transitionFilter.isPOSFiltered(),gdp.oracle.canDoNonProj());
	}

	@Override
	public TransitionSystem getTS() {
		throw new Error("todo"); //TODO need to fix this
	}	
	
}
