package ims.perceptron;

public abstract class AbstractLong2IntHash implements Long2IntInterface {
	private static final long serialVersionUID = 6316671241952533815L;

	protected final int size;
	protected transient int sf=0;
	
	protected AbstractLong2IntHash(int size){
		this.size=size;
	}
	
	@Override
	public int size() {
		return size;
	}

	@Override
	public void freeze() {
	}

	@Override
	public void unFreeze() {
	}

	@Override
	public boolean frozen() {
		return true;
	}

	@Override
	public int signFlips() {
		return sf;
	}
	
	public static class Long2IntAB extends AbstractLong2IntHash {
		private static final long serialVersionUID = 271090230266431329L;
		
		protected Long2IntAB(int size) {
			super(size);
		}

		@Override
		public int l2i(long l) {
			if(l<0){
				++sf;
				l=-l;
			}
			int a=(int) ((l<<33)>>>33); //Erase the 33 bits on the left. (== extract rightmost 31 bits)
			int b=(int) (l>>32);        //Extract the 32 bits on the left (== extract the leftmost 32 bits)
			int q=a^b;
			
			//So right now we ignore the 32nd bit since it might yield a negative q.
			//The other (pretty obvious option) is to do what Bernd does --
			//Don't care about negative q's, just take the modulo, and flip the sign if necessary
			//This way at least no information is discarded
			//Ok, this is in AB2	
			
//			System.out.printf("l: %65s, %2s\n",Long.toBinaryString(l),Long.signum(l));
//			System.out.printf("a: %65s, %2s\n",Integer.toBinaryString(a),Integer.signum(a));
//			System.out.printf("b: %33s                                , %2s\n",Integer.toBinaryString(b),Integer.signum(b));
//			System.out.printf("q: %65s, %2s\n",Integer.toBinaryString(q),Integer.signum(q));
//			System.out.println("    |------------------------------||------------------------------|");
			return q%size;
		} 
	}
	
	public static class Long2IntAB2 extends AbstractLong2IntHash {
		private static final long serialVersionUID = 2914447087876297723L;

		protected Long2IntAB2(int size) {
			super(size);
		}

		@Override
		public int l2i(long l) {
			if(l<0){
				++sf;
				l=-l;
			}
			int a=(int) ((l<<32)>>32); //Erase the 32 bits on the left. (== extract rightmost 32 bits)
			int b=(int) (l>>32);       //Extract the 32 bits on the left (== extract the leftmost 32 bits)
			int q=(a^b)%size;

			return q<0? -q : q;
		} 
	}
	
	public static class Long2IntABFast extends AbstractLong2IntHash {
		private static final long serialVersionUID = 3319487909346062693L;

		protected Long2IntABFast(int size) {
			super(size);
		}

		@Override
		public int l2i(long l) {
			int a=(int) ((l<<32)>>32); //Erase the 32 bits on the left. (== extract rightmost 32 bits)
			int b=(int) (l>>32);       //Extract the 32 bits on the left (== extract the leftmost 32 bits)
			int q=a^b;
			q&=0x7fffffff; //sign change
			return q%size;
		}
		
	}
	
	public static class Long2IntRBG extends AbstractLong2IntHash {
		private static final long serialVersionUID = 2842088055084781327L;

		protected Long2IntRBG(int size) {
			super(size);
		}

		@Override
		public int l2i(long l) {
	    	l = (l ^ (l&0xffffffff00000000L) >>> 32)*31;
	    	if(l<0){
	    		++sf;
	    		l=-l;
	    	}
	    	return (int)(l % size);
		}
		
	}
	
	public static class Long2IntModulo extends AbstractLong2IntHash {
		private static final long serialVersionUID = 3427006567836474238L;

		protected Long2IntModulo(int size) {
			super(size);
		}

		@Override
		public int l2i(long l) {
			return (int) (l<0?-(l%size):(l%size));
		}
		
	}
	
}
