package ims.tag;

import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.tag.features.AbstractFeatureSet;
import ims.tag.features.FeatureScorer;
import ims.tag.features.Potentials;

public class MultiPassDecoder implements IDecoder {
	private static final long serialVersionUID = 8850960940028322293L;
	
	private final FeatureScorer featureScorer;
	private final int passes;
	
	int totalParses=0;
	int totalPasses=0;
	int brokenPassLoops=0;
	
	public MultiPassDecoder(FeatureScorer featureScorer,int passes){
		this.featureScorer=featureScorer;
		this.passes=passes;
	}
	
	@Override
	public int[] decodeInstance(TagInstance instance, TagSymbolTable symTab) {
		++totalParses;
		Potentials potentials=featureScorer.getPotentials(instance);
		TagContext context=new TagContext(instance.forms.length);

		for(int idx=1;idx<instance.forms.length;++idx){
			int tag=potentials.bestPotential(idx);
			context.setTag(idx, tag);
		}

		int[] r=new int[instance.forms.length];
		
		for(int p=0;p<passes;++p){
			++totalPasses;
			int changes=0;
			for(int idx=1;idx<r.length;++idx){
				double[] contextScores=featureScorer.scoreNonLocal(idx, instance, context);
				int bestTag=-1;
				double bestTagScore=-99999.d;
				for(int vT:symTab.tagIdx){
					double potentialScore=potentials.getPotential(idx, vT);
					double contextScore=contextScores[vT];
					double score=potentialScore+contextScore;
					if(score>bestTagScore){
						bestTag=vT;
						bestTagScore=score;
					}
				}
				if(bestTag!=r[idx]){
					++changes;
					r[idx]=bestTag;
					context.setTag(idx, bestTag);
				}
			}
			if(changes==0){
				brokenPassLoops++;
				break;
			}
		}
		
		return r;
	}

	@Override
	public double recoverFeaturesAndFillFV(int[] tags, TagInstance inst, AbstractFeatureSet fs, IFV<FV3> fv, FeatureScorer featureScorer) {
		TagContext tagContext=new TagContext(tags);
		fv.clear();
		double score=0.d;
		for(int i=1;i<tags.length;++i){
			score+=fs.recoverFeaturesAndFillFV(inst, tagContext, i, tags[i], featureScorer, fv);
		}
		return score;
	}

}
