package ims.tag;

import ims.io.Format;
import ims.perceptron.Long2IntFactory.Long2IntImpl;
import ims.perceptron.Regularizer.RegularizerTypes;
import ims.tag.SamplingDecoder.IAcceptor;
import ims.tag.features.AbstractFeatureSet.FS;
import ims.trans.Options.ShuffleStrategy;
import ims.util.DBO;
import ims.util.ThreadPoolSingleton;
import ims.util.Util;

import java.io.File;
import java.io.OutputStream;
import java.util.Date;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

public class TagOptions {

	public final Date optionsCreate=new Date();
	
	@Option(name="-train",aliases="--train",usage="Train file")
	public File train;
	
	@Option(name="-out",aliases="--output",usage="Output file")
	public File out;

	@Option(name="-test",aliases="--test",usage="Test file")
	public File test;

	@Option(name="-gold")
	public File gold;

	@Option(name="-testEveryIter",aliases="-testEvery",usage="Test on test and evaluate on gold after every iteration")
	public boolean testEveryIter=false;

	@Option(name="-model",usage="Model file")
	public void setModel(String modelstr){
		if(modelstr.equalsIgnoreCase("null"))
			this.model=null;
		else
			this.model=new File(modelstr);
	}
	public File model=null;

	@Option(name="-iter",aliases="--iterations",usage="number of iterations for perceptron")
	public int iterations=10;
	
	@Option(name="-l2i")
	public Long2IntImpl l2iImpl=Long2IntImpl.Bohnet;
	
	@Option(name="-hsize")
	public int parameterSize=86028157; //this is a prime
	
	@Option(name="-reg",aliases="--regularizer",usage="which PA regularizer to use")
	public RegularizerTypes regularizerType=RegularizerTypes.Perceptron;

	public static boolean useAveraging=true;
	@Option(name="-noavg",aliases="--noAveraging",usage="Don't use averaging")
	public void setNoAvg(Boolean b){
		useAveraging=false;
	}

	@Option(name="-count",usage="maximum number of training instances")
	public int count=Integer.MAX_VALUE-1;

	@Option(name="-format",usage="set input format")
	public static Format format=Format.CoNLL06;

	@Option(name="-sinkSize")
	public static int sinkSize=1000;
	
	@Option(name="-affixLength")
	public int affixLength=4;

	@Option(name="-fs")
	public FS fs=FS.Local;
	
	@Option(name="-shuffle",aliases="-shuffleStrategy",usage="when to shuffle")
	public static ShuffleStrategy shuffleStrategy=ShuffleStrategy.Every;

	@Option(name="-passes")
	public static int passes=5;

	@Option(name="-n")
	public static int n=2;

	@Option(name="-greedyRestarts")
	public static int greedyRestarts=50;

	@Option(name="-randomRestarts")
	public static int randomRestarts=50;
	
	@Option(name="-resamplings")
	public static int resamplings=50;
	
	@Option(name="-tagChanges")
	public static int tagChanges=1;

	@Option(name="-entropyPropose")
	public static boolean entropyPropose=false;
	
	@Option(name="-coinFlipP")
	public static double coinFlipAcceptorP;

	@Option(name="-sampleAcceptor")
	public static IAcceptor.Type sampleAcceptor=IAcceptor.Type.OnlyIncrease;
	
	@Option(name="-debug")
	public static boolean DEBUG=false;

	@Option(name="-shuffleSeed")
	public long shuffleSeed=131716811161380l;


	public static enum Decoder { Greedy, Collins, Sampler }
	
	@Option(name="-decoder")
	public Decoder decoder=Decoder.Greedy;
	
	public TagOptions(String[] args) {
		CmdLineParser clp=new CmdLineParser(this);
		try {
			clp.parseArgument(args);
		} catch (CmdLineException e) {
			e.printStackTrace();
			System.err.println("Failed to parse cmd line, aborting.");
			System.err.println(e.getMessage());
			System.err.println();
			System.err.println("Valid options are:");
			printUsage(System.err);
			System.exit(1);
		}
	}

	private TagOptions(){}
	public static void printUsage(OutputStream out){
		CmdLineParser clp=new CmdLineParser(new TagOptions());
		clp.printUsage(out);
	}	
	public static void main(String[] args){
		printUsage(System.out);
	}
	
	public void done(){
		DBO.println("done() at "+new Date());
		DBO.println("Total time "+Util.insertCommas(System.currentTimeMillis()-optionsCreate.getTime()));
		ThreadPoolSingleton.shutdown();
		DBO.close();
	}
}
