package ims.tag.features;

import java.io.Serializable;

import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.IFV;
import ims.tag.ITagContext;
import ims.tag.TagContext;
import ims.tag.TagInstance;
import ims.tag.TagOptions;
import ims.util.ArrayFunctions;

public abstract class AbstractFeatureSet implements Serializable {
	
	private static final long serialVersionUID = 8816658649794647158L;
	
	public static final int NONE=AbstractSymbolMapping.NONE_IDX;
	
	public static enum FS { Local, Collins, NLinear }
	
	public static AbstractFeatureSet getFS(FS fs,TagSymbolTable symTab){
		switch(fs){
		case Local:   return new LocalFeatureSet(symTab);
		case Collins: return new CollinsFeatureSet(symTab);
		case NLinear: return new NLinearContextFeatureSet(symTab,TagOptions.n);
		default: throw new Error("not implemented");
		}
	}
	
	
	public final long[] sink;
	
	protected final int
		bF,
		bT,
		bA;
	
	protected AbstractFeatureSet(TagSymbolTable symTab){
		this.sink=new long[TagOptions.sinkSize];
		this.bF=symTab.forms.getBits();
		this.bA=symTab.affixes.getBits();
		this.bT=symTab.tags.getBits();
	}
	
	//These return pure features, tag of target not included
	public abstract int extractLocalFeatures(int idx,TagInstance inst);
	public abstract int extractNonLocalFeatures(int idx,TagInstance inst,ITagContext context);

	public double recoverFeaturesAndFillFV(TagInstance inst, TagContext context, int idx, int tag, FeatureScorer featureScorer, IFV<?> fv) {
		int m=extractLocalFeatures(idx,inst);
		ArrayFunctions.shiftAllLeftByBits(sink, bT, m);
		for(int i=0;i<m;++i){
			sink[i]|=tag;
			int y=featureScorer.l2i.l2i(sink[i]);
			fv.add(y);
		}
		double score=featureScorer.scoreFeatures(sink, m);
		
		int n=extractNonLocalFeatures(idx,inst,context);
		ArrayFunctions.shiftAllLeftByBits(sink, bT, n);
		for(int i=0;i<n;++i){
			sink[i]|=tag;
			int y=featureScorer.l2i.l2i(sink[i]);
			fv.add(y);
		}

		score+=featureScorer.scoreFeatures(sink, n);
		
		return score;
	}
	
}
