package ims.trans;

import java.io.IOException;
import java.util.Arrays;

import ims.data.NBestCoNLL09Sentence;
import ims.data.Parse;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.util.DBO;
import ims.util.ZipUtil;

public class RescoreNBest {

	public static void main(String[] args) throws IOException, ClassNotFoundException{
		Options options=new Options(args);
		IOracle oracle=OracleFactory.getOracle(options);
				
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		Decoder d=new Decoder(m);
		
		SentenceReader<NBestCoNLL09Sentence> reader=ReaderWriterFactory.getReader(options.test, NBestCoNLL09Sentence.class,options.reversed);
		SentenceWriter writer=ReaderWriterFactory.getWriter(options.out, options.reversed);

		DBO.println("Rescoring.");
		int i=0;
		int erase=0;
		int fail=0;
		int parses=0;
		DBO.printNoPrefix("Sent: ");
		for(NBestCoNLL09Sentence sent:reader){
			int[] oldGHeads=Arrays.copyOf(sent.getPredHeads(),sent.getLength());
			String[] oldGLbls=Arrays.copyOf(sent.getPredLabels(),sent.getLength());
			parses+=sent.getParses().size();
			for(Parse parse:sent.getParses()){
				int[] heads=parse.heads;
				String[] labels=parse.lbls;
				sent.setGoldHeads(heads);
				sent.setGoldLabels(labels);
				try {
					State state=d.decodeGold(sent, m.beamSize, oracle);
					parse.properties.put(Parse.PARSE_PROPERTY_MAP_IMSTRANS_PARSE_RESCORE, Double.toString(state.score()));
				} catch(IllegalArgumentException e) { //This happens if we need a transition that is not defined. Then we cannot define the gold score according to the parser.
					parse.properties.put(Parse.PARSE_PROPERTY_MAP_IMSTRANS_PARSE_RESCORE, "n/a");
					++fail;
				}
				
			}
			if(i++%111==0){
				erase=DBO.eraseAndPrint(erase, i);
			}
			sent.setGoldHeads(oldGHeads);
			sent.setGoldLabels(oldGLbls);
			writer.write(sent);
		}
		writer.close();
		DBO.println();
		DBO.println("Done.");
		DBO.println();
		DBO.println("Failed: "+String.format("%d / %d = %5.2f%%\n", fail,parses,(100.*fail/parses)));
	}
	
}
