package ims.trans;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.IOracle;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.util.DBO;

import java.util.List;

public class SamplingTrainer extends AbstractTrainer {
	
	private final SamplingDecoder sd;
	
	public SamplingTrainer(Options options, SymbolTable symTab, IOracle oracle,	ITransitionFilter transitionFilter, List<Instance> instances, FeatureExtractor featureExtractor, Long2IntInterface l2i,	ParametersFloat parameters, LossFunction lossFunction, InstanceMapping instanceMapping) {
		super(options, symTab, oracle, transitionFilter, featureExtractor, l2i, parameters, lossFunction, instanceMapping);
		SamplingStoppingCriterion stopper=SamplingStoppingCriterion.get(options.stopC, options.sampleK);
		this.sd=new SamplingDecoder(this.parameters.parameters,this.l2i,this.symTab,this.transitionFilter,this.featureExtractor,options.mergeEquivalentStates,options.labeled,this.instanceMapping,stopper);
	}

	@Override
	protected int doIteration(int itemCount,List<Instance> instances) {
		long start=System.currentTimeMillis();
		IFV<FV3> goldFV=new FV3();
		IFV<FV3> predFV=new FV3();
		int erase=0;
		int badSearch=0;
		int correct=0;
		double accLoss=0;
		int isCount=0;
		Bookkeeping b=new Bookkeeping();
		DBO.printNoPrefix("Sent: ");
		for(Instance is:instances){
			++isCount;
			++itemCount;
			oracle.initSentence(is);
			if(isCount%111==1)
				erase=printInfo(erase,b,accLoss,isCount,badSearch,correct,System.currentTimeMillis()-start);
			State gold=decodeGold(is);
			State pred=sd.findBestSample(is);

			b.goldTrans+=gold.totalTransitions;
			b.predTrans+=pred.totalTransitions;
			
			b.goldSwaps+=gold.countSwaps();
			b.predSwaps+=pred.countSwaps();
			
			
			boolean isCorrect=checkCorrectTree(pred, is);
			if(!isCorrect){
				if(gold.score()>pred.score()){ //Then we had bad luck in sampling
					++badSearch;
				} else {
					accLoss+=super.doUpdate(itemCount, goldFV, predFV, is, pred, gold);
				}
			} else {
				++correct;
			}
		}
		erase=printInfo(erase,b,accLoss,isCount,badSearch,correct,System.currentTimeMillis()-start);
		DBO.println();
		return itemCount;
	}

	private int printInfo(int erase, Bookkeeping b, double accLoss,	int isCount, int badSearch, int correct,long time) {
		double gt=(double)b.goldTrans/isCount;
		double pt=(double)b.predTrans/isCount;
		double gs=(double)b.goldSwaps/isCount;
		double ps=(double)b.predSwaps/isCount;
		double l=(double)accLoss/isCount;
		long t=time/isCount;
		String info=String.format("%-5d | c: %-5d, bs: %-5d, t: %dms | l: %5.2f |gt: %6.3f  pt: %6.3f |gs: %6.3f  ps: %6.3f",isCount,correct,badSearch,t,l,gt,pt,gs,ps);
		int e=DBO.eraseAndPrint(erase, info);
		return e;
	}

	private State decodeGold(Instance is) {
		//Do it with static sequence for now, XXX: maybe change later to sample with dyn oracle
		State s=State.getInitialState(is.form.length, super.options.labeled, null);
		for(int t:is.transSequence){
			ITransition tr=symTab.transition2label[t];
			double[] scores=sd.scoreTransitions(s, is, new int[]{t});
			s=tr.apply(s,scores[0]);
		}
		return s;
	}

}
