package ims.trans.doc;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import ims.data.Document;
import ims.data.ISentence;
import ims.io.DocumentWriter;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.trans.AbstractDecoder;
import ims.trans.AbstractTrainer;
import ims.trans.Decoder;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.State;
import ims.trans.util.Eval;
import ims.util.Accuracy;
import ims.util.DBO;
import ims.util.Pair;
import ims.util.PrecisionRecall;
import ims.util.Util;
import ims.util.ZipUtil;

public class Test implements AbstractTrainer.Tester {
	
	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		DBO.println("Loading model");
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		DBO.println("Done.");
		
		testModel(m,options);
		
		if(options.gold!=null){
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(options.out, options.gold, Options.format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		}
		
		options.done();
	}
	
	
	private static void testModel(Model m,Options options) throws IOException{
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.test, Options.format);
		DocumentWriter    writer=new DocumentWriter(options.out);
		testModel(m,reader,writer);
		writer.close();
	}
	
	private static void testModel(Model m, SentenceReader<?> reader, DocumentWriter writer) throws IOException {
		Decoder d=new Decoder(m);
		testDecoder(d,reader,writer);
	}


	//	public static <T extends ISentence> void testModel(Model m,SentenceReader<T> reader,DocumentWriter writer,int beamSize) throws IOException{
	public static <T extends ISentence> void testDecoder(AbstractDecoder d,SentenceReader<T> reader,DocumentWriter writer) throws IOException{
//		Decoder decoder=new Decoder(m);
		long start=System.currentTimeMillis();
		int erase=0;
		int sentCount=0;
		DBO.printNoPrefix("Sentence: ");
		for(T sent:reader){
//			Instance inst=m.instanceMapping.createTestInstance(sent, m.symTab);
//			Instance inst=m.instanceMapping.createTestInstance(d,m.symTab);
//			Instance inst=d.instanceMapping.createTestInstance(sent, d.symTab);
			State s=d.parse(sent);
//			if(beamSize>1)
//				s=d.decodeBeam(inst, beamSize);
//			else
//				s=d.decodeGreedy(inst);
			
			
			
			Pair<int[],String[]> tree=d.state2tree(s);
			List<Integer> sentBoundaries=s.getSentenceBoundaries();
//			TIntList sentBoundaries=s.getSentenceBoundaries();
			sent.setPredHeads(tree.getLeft());
			sent.setPredLabels(tree.getRight());

			List<T> outSents=new ArrayList<>();
			List<String> sIDs=new ArrayList<>();

			int sId=1;
			int b=1;
			int tokOffset=0;
//			for(TIntIterator it=sentBoundaries.iterator();it.hasNext();){
//				int e=it.next();
			for(Integer e:sentBoundaries){
				T p=(T) sent.getSubSentence(b,e,tokOffset);
				b=e;
				outSents.add(p);
				sIDs.add(Integer.toString(sId++));
				tokOffset+=p.getLength()-1;
			}
			//Then we need to get the last sentence too
			T p=(T)sent.getSubSentence(b,sent.getLength(),tokOffset);
			outSents.add(p);
			sIDs.add(Integer.toString(sId++));
			
			String dId=sent.getDocID();
			
			Document<T> doc=new Document<T>(outSents, sIDs, dId);
			writer.write(doc);
			++sentCount;
			erase=DBO.eraseAndPrint(erase, String.format("%-5d   (%d/s)", sentCount,(System.currentTimeMillis()-start)/sentCount));
		}
		DBO.println();
	}


	@Override
	public void evalTest(File goldFile, File predFile) {
		Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(predFile, goldFile, Options.format);
		PrecisionRecall pr=Eval.evalSentBoundaries(predFile, goldFile);
		Accuracy[] em=Eval.getExactMatchSentBoundaries(predFile, goldFile);
		System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		System.out.println("SB:");
		System.out.println(pr.getPRFString());
		System.out.println("Exact matches (S,S+U,S+U+L)");
		System.out.println(Util.joinStrings("    ",em));
	}

	@Override
	public void applyDecoder(AbstractDecoder decoder, SentenceReader<?> reader, File output, boolean reversed) throws IOException {
		DocumentWriter writer=new DocumentWriter(output);
		testDecoder(decoder,reader,writer);
	}
	
}
