package ims.trans.features;

import ims.data.Instance;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.trans.FeatureExtractor;
import ims.trans.State;

public class FS_Default extends FeatureExtractor{
	private static final long serialVersionUID = -6434318953730936842L;

	public FS_Default(SymbolTable symTab) {
		super(symTab);
	}

	public int extractFeatures(State s,long[] sink,Instance inst){
		
		//Stack
		int s0=s.s0;
		int s1,s2,s3;
		s1=s2=s3=-1;
		int s0f,s0t,s1f,s1t,s2f,s2t,s3f,s3t,s1ht,s1hf;
		int s0lp,s0rp,b0lp,b0rp,s1rp,s1lp,s1lf,s1rf;
		s0f=s0t=s1f=s1t=s2f=s2t=s3f=s3t=s0lp=s0rp=b0lp=b0rp=s1rp=s1lp=s1ht=s1hf=s1lf=s1rf=NONE;
		
		int s1l,s1ll,s1rl,s0ll; //Labels
		s1l=s1ll=s1rl=s0ll=NONE;
		
		if(s0>=0){  //s0 a token?
			s0f=inst.form[s0];
			s0t=inst.tag[s0];
			if(s.tail!=null){ //s0 has other item on the stack below?
				s1=s.tail.s0;
			}
			if(s.s0L>0){ 
				s0lp=inst.tag[s.s0L];
//				s0ll=s.s0Llbl;
				s0ll=s.getLabel(s.s0L);
			}
			if(s.s0R>0){
				s0rp=inst.tag[s.s0R];
			}
		}
		if(s1>=0){ //s1 is a token
			s1f=inst.form[s1];
			s1t=inst.tag[s1];
//			s1l=s.tail.s0lbl;
			s1l=s.getLabel(s.tail.s0);
			if(s.tail.tail!=null) //s1 has other item on the stack below{
				s2=s.tail.tail.s0;
			int s1h=s.getHead(s1);
			if(s1h>0){ //This should really by =>0 since 0 can be the head. Also, we should change this to s0h (I misinterpreted the Zhang and Nivre features) 
				s1ht=inst.tag[s1h];
				s1hf=inst.form[s1h];
			}
			if(s.tail.s0R>0){
				s1rp=inst.tag[s.tail.s0R];
				s1rf=inst.form[s.tail.s0R];
//				s1rl=s.tail.s0Rlbl;
				s1rl=s.getLabel(s.tail.s0R);
			}
			if(s.tail.s0L>0){
				s1lp=inst.tag[s.tail.s0L];
				s1lf=inst.tag[s.tail.s0L];
//				s1ll=s.tail.s0Llbl;
				s1ll=s.getLabel(s.tail.s0L);
			}
		}
		if(s2>=0){
			s2f=inst.form[s2];
			s2t=inst.form[s2];
			if(s.tail.tail.tail!=null)
				s3=s.tail.tail.tail.s0;
		}
		if(s3>=0){
			s3f=inst.form[s3];
			s3t=inst.tag[s3];
//			if(s.tail.tail.tail.tail!=null)
		}
		
		//Buffer
		int b0=s.buf.get(0);
		int b1=s.buf.get(1);
		int b2=s.buf.get(2);		
		
		int b0f=b0!=-1?inst.form[b0]:AbstractSymbolMapping.NONE_IDX;
		int b0t=b0!=-1?inst.tag[b0]:AbstractSymbolMapping.NONE_IDX;
		
		int b1f,b1t,b2f,b2t;
		b1f=b1t=b2f=b2t=NONE;
		if(b1>0){
			b1f=inst.form[b1];
			b1t=inst.form[b1];
		}
		if(b2>0){
			b2f=inst.form[b2];
			b2t=inst.tag[b2];
		}
		
		
		
		int t=1; //Template index
		int i=0; //Where to put in sink
		
		{	//Own features for swap (hopefully reducing swaps
			int lastTrans=AbstractSymbolMapping.NONE_IDX;
			int lastLastTrans=AbstractSymbolMapping.NONE_IDX;
			if(s.lastAction!=null){
				lastTrans=s.lastAction.getId();
				if(s.previous.lastAction!=null)
					lastLastTrans=s.previous.lastAction.getId();
			}
			boolean hasSwapped=s.buf.hasSwappedItems();
			
			sink[i]=t; sink[i]<<=bA; sink[i]|=lastTrans; ++t; ++i;
			sink[i]=t; sink[i]<<=bA; sink[i]|=lastTrans; sink[i]<<=bA; sink[i]|=lastLastTrans; ++t; ++i;
			sink[i]=t; sink[i]<<=1;  sink[i]|=(hasSwapped?1:0); ++t;++i;
		}
		
		{   //Table 1 from Zhang and Nivre, all is shifted for the different transition system though.
			//the comments use the notation from Zhang and Nivre, but the extraction uses
			//the two topmost stack items for stack top and buffer front
			
			//Singles
			sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; ++t; ++i;                             //s0w
			sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;                             //s0p
			sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i; //s0wp

			sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i;                             //b0w
			sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i;                             //b0p
			sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //b0wp

			sink[i]=t; sink[i]<<=bF; sink[i]|=b0f; ++t; ++i;                             //b1w
			sink[i]=t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;                             //b1p
			sink[i]=t; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i; //b1wp

			sink[i]=t; sink[i]<<=bF; sink[i]|=b1f; ++t; ++i;                             //b2w
			sink[i]=t; sink[i]<<=bT; sink[i]|=b1t; ++t; ++i;                             //b2p
			sink[i]=t; sink[i]<<=bF; sink[i]|=b1f; sink[i]<<=bT; sink[i]|=b1t; ++t; ++i; //b2wp 

			//Word pairs
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i; //s0wp b0w
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0wp b0p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0p b0wp
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0w b0wp
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0wp b0wp

			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i;	//s0w b0w
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i;	//s0p b0p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;	//b0p b1p

			//Three words
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bT;sink[i]|=b1t; ++t; ++i;	//b0p b1p b2p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=b0t; ++t; ++i; //s0p b0p b1p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1ht;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i;//s0ht s0t b0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1lp;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s0p s0lp b0p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1rp;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s0p s0rp b0p
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0lp; ++t; ++i; //s0p b0p b0lp
		}
		
		{ //Zhang and Nivre distance features
			
			int distBucketed=getIntBucket(s1>=0?Math.abs(s0-s1):-1);
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s1f; ++t; ++i; //s0w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s1t; ++t; ++i; //s0t d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s0f; ++t; ++i; //b0w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //b0p d
			
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s1f;sink[i]<<=bF;sink[i]=s0f; ++t; ++i; //s0w b0w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]=s0t; ++t; ++i; //s0t b0t d
		}
		
		{ //Zhang and Nivre valency features -- skip for now, since the graph-structured stack doesn't make this all that easy
			int s1vR,s1vL,s0vR,s0vL;
			s1vR=s1vL=s0vR=s0vL=AbstractSymbolMapping.NONE_IDX;
		}
		
		{ //Zhang and Nivre unigrams
			sink[i]=t;sink[i]<<=bF;sink[i]|=s1hf; ++t; ++i; //s0hw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1ht; ++t; ++i; //s0ht
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1l;  ++t; ++i; //s0l  //label

			sink[i]=t;sink[i]<<=bF;sink[i]|=s1lf; ++t; ++i; //s0lw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1lp; ++t; ++i; //s0lt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1ll; ++t; ++i; //s0ll //label on leftmost
			
			sink[i]=t;sink[i]<<=bF;sink[i]|=s1rf; ++t; ++i; //s0rw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1rp; ++t; ++i; //s0rt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1rl; ++t; ++i; //s0rl //label on rightmost
			
			sink[i]=t;sink[i]<<=bF;sink[i]|=s0f;  ++t; ++i; //b0w
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;  ++t; ++i; //b0t;
			sink[i]=t;sink[i]<<=bL;sink[i]|=s0ll; ++t; ++i; //b0ll //label on leftmost
//			N0l l; //N/A
		}

		{ //Zhang and Nivre third-order
//			S0h2 w; S0h2 p; S0h l; S0l2 w; S0l2 p; S0l2 l;
//			S0r2 w; S0r2 p; S0r2 l; N0l2 w; N0l2 p; N0l2 l;
//			S0 pS0l pS0l2 p; S0 pS0r pS0r2 p;
//			S0 pS0h pS0h2 p; N0 pN0l pN0l2 p;

		}
		
		return i;
	}
	

}
