package ims.trans.features;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;
import ims.data.Instance;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.trans.FeatureExtractor;
import ims.trans.Options;
import ims.trans.State;
import ims.trans.transitions.ITransition;
import ims.util.DBO;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

public class FS_XDStandard extends FeatureExtractor {
	private static final long serialVersionUID = -6796860041944069747L;

	final boolean useB0Context;
	final boolean useStag;
	final boolean useStack;
	final boolean useShapes;
	final TIntSet verbPosTags=new TIntHashSet();

	final boolean useSentenceSizeSpan;
	final boolean useStackSize;
	final boolean useSubCats;
	
	public FS_XDStandard(SymbolTable symTab,Options options){
		super(symTab);
		this.useStag=options.useStag;
		this.useStack=options.useStack;
		this.useB0Context=options.b0context;
		this.useShapes=options.useShape;
		String[] tags=symTab.tags.getInverted();
		//Populate verb tags
		for(int i=0;i<tags.length;++i){
			if(tags[i]!=null && tags[i].startsWith("V"))
				verbPosTags.add(i);
		}
		
		
		useSentenceSizeSpan=Options.x0;
		useStackSize=Options.x1;
		useSubCats=Options.x2;
	}

	private static final int INIT_T=1;

	private static final float BIAS_INIT_ARC=0.001f;
	private static final float BIAS_INIT_SH=0.0001f;
	private static final float BIAS_INIT_SW=0.00001f;



	@Override
	public void initWeights(SymbolTable symTab,ParametersFloat parameters,Long2IntInterface l2i) {
		DBO.println("Initializing bias features");
		Set<Integer> seen=new HashSet<Integer>();
		long l=getBias(INIT_T);
		for(int i=0;i<symTab.transition2label.length;++i){
			ITransition t=symTab.transition2label[i];
			if(t==null)
				continue;
			long k=l;
			k<<=bA;
			k|=t.getId();
			int h=l2i.l2i(k);
			if(seen.contains(h)){
				throw new Error("Hash collision during feature weight init -- try changing hash size");
			}
			seen.add(h);
			if(t.isArcTransition()){
				parameters.parameters[h]=BIAS_INIT_ARC;
			} else if(t.getTR()==TR.SW){
				parameters.parameters[h]=BIAS_INIT_SW;
			} else if(t.getTR()==TR.SH){
				parameters.parameters[h]=BIAS_INIT_SH;
			} else {
				throw new Error("this should not happen -- unknown transition?");
			}
		}
	}

	private long getBias(int t){
		long l=t;
		l<<=7;
		l|=0xa;
		return l;
	}

	
	//TODO: features
	// - Bucket sentence-initial character: Uppercase, Lowercase, numerical, punct, (other?)
	// - type -> sentence begin ratio mined on unlabeled data for unsupervised tokenization (schmid, kiss), fire feature on b0
	// - stacksize
	// - size of span that might be closed (i.e., distance from b0 to last sentence break)
	// - # verbs in stack OR # of verbs from last SB?
	// - features on topmost verb in stack (subcat frame?)
	// - subcat frames in general?
	// - add information about how many root dependents within current sentence. for this we have to extend the state and track the previously predicted sentence boundary too. this could potentially help avoiding putting multiple root dependents.
	
	public int extractFeatures(State s,long[] sink,Instance inst){
		int t=INIT_T; //Template index
		int i=0; //Where to put in sink

		//I think we need these features even if the action is unique
		//Yes, let's forget about this and return a bit later. Returning this early seems harmful
		//		if(s.tail==null){ //If we only have root on stack we have to shift anyway, can just as well minimize feature extraction
		//			sink[i]=t; sink[i]<<=7; sink[i]|=0xf; ++i; //must shift feature
		//			return i;
		//		}
		//		//I had this before since it feels stupid to extract features on the very last RA action, but that might not be the case with the beam, so let's take it out again
		////		else if(s.tail.s0==0 && s.bufferSize()==0){ //must RA feature
		////			sink[i]=t; sink[i]<<=7; sink[i]|=0x7; ++i;
		////			return i;
		////		} 

		sink[i]=getBias(t); ++i; ++t;

		//Stack
		final int s0=s.s0;

		final int s0f=inst.form[s0];
		final int s0t=inst.tag[s0];
		final int[] s0m=inst.feat[s0];

		final int s1=s.tail==null?-1:s.tail.s0;

		final int s1f;
		final int s1t;
		final int[] s1m;

		final int s0lp,s0rp,s0lf,s0rf,s0ll,s0rl; //leftmost and rightmost deps of s0
		final int s1lp,s1rp,s1lf,s1rf,s1ll,s1rl; //leftmost and rightmost deps of s1

		final int[] s0lm,s0rm,s1lm,s1rm;         //left and right morphology of s0 and s1


		final int s2t,s2f;	

		if(s.s0L>0){
			s0lf=inst.form[s.s0L];
			s0lp=inst.tag[s.s0L];
			s0ll=s.getLabel(s.s0L);
			s0lm=inst.feat[s.s0L];
		} else {
			s0lf=NONE;
			s0lp=NONE;
			s0ll=NONE;
			s0lm=NONE_ARR;
		}
		if(s.s0R>0){
			s0rf=inst.form[s.s0R];
			s0rp=inst.tag[s.s0R];
			s0rl=s.getLabel(s.s0R);
			s0rm=inst.feat[s.s0R];
		} else {
			s0rf=NONE;
			s0rp=NONE;
			s0rl=NONE;
			s0rm=NONE_ARR;
		}

		if(s1>=0){
			s1f=inst.form[s1];
			s1t=inst.tag[s1];
			s1m=inst.feat[s1];
			if(s.tail.s0L>0){
				s1lp=inst.tag[s.tail.s0L];
				s1lf=inst.form[s.tail.s0L];
				s1ll=s.getLabel(s.tail.s0L);
				s1lm=inst.feat[s.tail.s0L];
			} else {
				s1lp=NONE;
				s1lf=NONE;
				s1ll=NONE;
				s1lm=NONE_ARR;
			}
			if(s.tail.s0R>0){
				s1rp=inst.tag[s.tail.s0R];
				s1rf=inst.form[s.tail.s0R];
				s1rl=s.getLabel(s.tail.s0R);
				s1rm=inst.feat[s.tail.s0R];
			} else {
				s1rp=NONE;
				s1rf=NONE;
				s1rl=NONE;
				s1rm=NONE_ARR;
			}
			final int s2=s.tail.tail==null?-1:s.tail.tail.s0;
			if(s2>-1){
				s2t=inst.tag[s2];
				s2f=inst.form[s2];
			} else {
				s2t=NONE;
				s2f=NONE;
			}
		} else {
			s1f=NONE;
			s1t=NONE;
			s1m=NONE_ARR;
			s1lp=NONE;
			s1rp=NONE;
			s1lf=NONE;
			s1rf=NONE;
			s1ll=NONE;
			s1rl=NONE;
			s1lm=NONE_ARR;
			s1rm=NONE_ARR;
			s2f=NONE;
			s2t=NONE;
		}

		//Buffer
		int b0=s.buf.get(0);

		final int b0f;
		final int b0t;
		final int b1f;
		final int b1t;


		final int b2t;
		final int b2f;
		
		final int b0shape;
		final int b0fcs;
		final int b0m1shape;

		if(b0>0){
			b0f=inst.form[b0];
			b0t=inst.tag[b0];
			
			b0shape=inst.shape[b0];
			b0fcs=inst.firstCharShape[b0];
			b0m1shape=inst.shape[b0-1];
			int b1=s.buf.get(1);
			if(b1>0){
				b1f=inst.form[b1];
				b1t=inst.tag[b1];
				int b2=s.buf.get(2);
				if(b2>0){
					b2t=inst.tag[b2];
					b2f=inst.form[b2];
				} else {
					b2t=NONE;
					b2f=NONE;
				}
			} else {
				b1f=NONE;
				b1t=NONE;
				b2f=NONE;
				b2t=NONE;
			}
		} else {
			b0f=NONE;
			b0t=NONE;
			b1f=NONE;
			b1t=NONE;
			b2f=NONE;
			b2t=NONE;
			b0shape=NONE;
			b0fcs=NONE;
			b0m1shape=NONE;
		}
		
		final int distToLastSB=getSentLengthBucket(-1);

		{   //Table 1 from Zhang and Nivre, all is shifted for the different transition system though.

			//Singles
			sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; ++t; ++i;                             //s1w
			sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;                             //s1t
			sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i; //s1wt

			sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i;                             //s0w
			sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i;                             //s0t
			sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0wt

			sink[i]=t; sink[i]<<=bF; sink[i]|=b0f; ++t; ++i;                             //b0w
			sink[i]=t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;                             //b0t
			sink[i]=t; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i; //b0wt

			sink[i]=t; sink[i]<<=bF; sink[i]|=b1f; ++t; ++i;                             //b1w
			sink[i]=t; sink[i]<<=bT; sink[i]|=b1t; ++t; ++i;                             //b1t
			sink[i]=t; sink[i]<<=bF; sink[i]|=b1f; sink[i]<<=bT; sink[i]|=b1t; ++t; ++i; //b1wt

			//Word pairs
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i; //s1wt s0w
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s1wt s0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s1t s0wt
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s1w s0wt
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s1wt s0wt

			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i;	//s1w s0w
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i;	//s1t s0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;	//s0t b0t

			//Three words
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bT;sink[i]|=b1t; ++t; ++i;	 //s0t b0t b1t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=b0t; ++t; ++i;  //s1t s0t b0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1lp;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s1t s1lt s0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1rp;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s1t s1rt s0t
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0lp; ++t; ++i; //s1t s0t s0lt
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0rp; ++t; ++i; //s1t s0t s0rt
		}
		//Morphology on s0 and s1
		for(int a=0;a<s0m.length;++a){
			for(int b=0;b<s1m.length;++b){
				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bM;sink[i]|=s0m[a];sink[i]<<=bM;sink[i]|=s1m[b]; ++i;
			}
		}
		++t;

		i=extractB0Context(inst,b0,b0t,t,i);
		t+=11;		
		
		//Some shapes
		if(useShapes){
			final int stackSize=getIntBucket(s.expensiveStackSize());
			final int s0shape=inst.shape[s0];
			final int s0lcs=inst.lastCharShape[s0];
			final int s0fcs=inst.firstCharShape[s0];
			final int b0lcs=b0>0?inst.lastCharShape[b0]:NONE;
			
			final int s0p1shape,s0p2shape,s0m1shape,s0m2shape;
			final int b0p1shape,b0p2shape,b0m2shape;
			final int b0p1lcs,b0p2lcs,b0m1lcs,b0m2lcs;
			final int s0p1lcs,s0p2lcs,s0m1lcs,s0m2lcs;
			final int b0m1t;
			
			final int lastTR=s.lastAction==null?NONE:s.lastAction.getId();
			
			if(b0>0 && b0+1<inst.form.length){
				b0p1shape=inst.shape[b0+1];
				b0p1lcs=inst.lastCharShape[b0+1];
				if(b0+2<inst.form.length){
					b0p2shape=inst.shape[b0+2];
					b0p2lcs=inst.lastCharShape[b0+2];
				} else {
					b0p2shape=NONE;
					b0p2lcs=NONE;					
				}
			} else {
				b0p1shape=NONE;
				b0p1lcs=NONE;
				b0p2shape=NONE;
				b0p2lcs=NONE;
			}
			
			if(b0-1>=0){
				b0m1lcs=inst.shape[b0-1];
				b0m1t=inst.tag[b0-1];
				if(b0-2>=0){
					b0m2shape=inst.shape[b0-2];
					b0m2lcs=inst.shape[b0-2];
				} else {
					b0m2shape=NONE;
					b0m2lcs=NONE;
				}
			} else {
				b0m1lcs=NONE;
				b0m2shape=NONE;
				b0m2lcs=NONE;
				b0m1t=NONE;
			}
			
			if(s0+1<inst.form.length){
				s0p1shape=inst.shape[s0+1];
				s0p1lcs=inst.lastCharShape[s0+1];
				if(s0+2<inst.form.length){
					s0p2shape=inst.shape[s0+2];
					s0p2lcs=inst.shape[s0+2];
				} else {
					s0p2shape=NONE;
					s0p2lcs=NONE;
				}
			} else {
				s0p1shape=NONE;
				s0p1lcs=NONE;
			}
			
			if(s0-1>=0){
				s0m1shape=inst.shape[s0-1];
				s0m1lcs=inst.lastCharShape[s0-1];
				if(s0-2>=0){
					s0m2shape=inst.shape[s0-2];
					s0m2lcs=inst.shape[s0-2];
				} else {
					s0m2shape=NONE;
					s0m2lcs=NONE;
				}
			} else {
				s0m1shape=NONE;
				s0m1lcs=NONE;
				s0m2shape=NONE;
				s0m2lcs=NONE;				
			}
			
			//Now make features
			//XXX baseline (only shapes)
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0p1shape; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0p1shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0p1shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0m1shape; ++i; ++t;
//			
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; ++i; ++t;
////			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0m1shape; ++i; ++t;
			//XXX end baseline
			
			//XXX baseline conjoined with last action
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bA; sink[i]|=lastTR; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0p1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bA; sink[i]|=lastTR; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0p1shape; sink[i]<<=bA; sink[i]|=lastTR; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0p1shape; ++i; ++t;
////			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0m1shape; ++i; ++t;
//			
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bA; sink[i]|=lastTR; ++i; ++t;
////			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bA; sink[i]|=lastTR; ++i; ++t;
////			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0m1shape; ++i; ++t;
			//XXX end baseline+last action
			
			//XXX baseline+stacksize
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bD; sink[i]|=stackSize; sink[i]<<=bD; sink[i]|=distToLastSB; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0p1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bD; sink[i]|=stackSize; sink[i]<<=bD; sink[i]|=distToLastSB; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0p1shape; sink[i]<<=bD; sink[i]|=stackSize; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0p1shape; ++i; ++t;
//			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=s0m1shape; ++i; ++t;
			
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bD; sink[i]|=stackSize; sink[i]<<=bD; sink[i]|=distToLastSB; ++i; ++t;
//			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bD; sink[i]|=stackSize; sink[i]<<=bD; sink[i]|=distToLastSB; ++i; ++t;
//			sink[i]=t; sink[i]<<=bCS; sink[i]|=b0m1lcs; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0m1shape; ++i; ++t;
			//XXX end baseline+stacksize
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=distToLastSB; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bSH; sink[i]|=b0m1shape; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=b0m1t; ++i; ++t;
			sink[i]=t; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bT; sink[i]|=b0m1t; ++i; ++t;
		}
		
		//HERE WE STOP IF s1 is undefined XXX
		if(s1==-1){
			return i;
		}
		
		final int stackSize;
		final int topMostVerb;
		final int topMostVerbDepth;
		final int topMostVerbPos;
		final int topMostVerbForm;
		{ //Here scan the stack for interesting stuff.
			int ss=0;
			int topVerb=-1;
			int topVerbDepth=-1;
			State c=s;
			while(c!=null){
				++ss;
				if(topVerb!=-1 && verbPosTags.contains(inst.tag[c.s0])){
					topVerb=c.s0;
					topVerbDepth=i;
				}
				c=c.tail;
			}
			topMostVerb=topVerb;
			topMostVerbPos=topVerb>0?inst.tag[topVerb]:NONE;
			topMostVerbForm=topVerb>0?inst.form[topVerb]:NONE;
			topMostVerbDepth=getStackSizeBucket(topVerbDepth);
			stackSize=getStackSizeBucket(ss);
		}

		
		if(useSentenceSizeSpan){
			i=addDistBucketSBFeatures(distToLastSB,i,t,s0t,s1t,s0f,s1f,b0t,b0f,b0shape,b0m1shape,b0fcs,3);
			if(useShapes)
				t+=10;
			else
				t+=6;
		}
		
		if(useStackSize){
			i=addDistBucketSBFeatures(distToLastSB,i,t,s0t,s1t,s0f,s1f,b0t,b0f,b0shape,b0m1shape,b0fcs,stackSize);
			if(useShapes)
				t+=10;
			else
				t+=6;
		}
		
		
		//XXX add this in later
		if(false){ //This is a combo between distToLast and stackSize. Not sure with what to bucket that.
			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		}
		
		if(useSubCats){
			final short[] leftLabels;
			final short[] rightLabels;
			final int targetWord=topMostVerb;
			final int tPos;
			final int tForm;
			if(targetWord<0){
				leftLabels=NONE_SHORT_ARR;
				rightLabels=NONE_SHORT_ARR;
				tPos=NONE;
				tForm=NONE;
			} else {
				tPos=inst.tag[targetWord];
				tForm=inst.form[targetWord];
				switch(s.leftLabels[targetWord].length){
				case 0:  leftLabels=NONE_SHORT_ARR; break;
				case 1:  leftLabels=new short[]{s.leftLabels[targetWord][0]}; break;
				default: leftLabels=Arrays.copyOf(s.leftLabels[targetWord],s.leftLabels[targetWord].length); Arrays.sort(leftLabels); break;
				}
				switch(s.rightLabels[targetWord].length){
				case 0:  rightLabels=NONE_SHORT_ARR; break;
				case 1:  rightLabels=new short[]{s.rightLabels[targetWord][0]}; break;
				default: rightLabels=Arrays.copyOf(s.rightLabels[targetWord],s.rightLabels[targetWord].length); Arrays.sort(rightLabels); break;
				}
			}
			
			if(leftLabels.length==1){ //left-left
				sink[i]=t;   sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bL; sink[i]|=leftLabels[0]; sink[i]<<=bL; sink[i]|=NONE; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
				sink[i]=t+1; sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bF; sink[i]|=tForm; sink[i]<<=bL; sink[i]|=leftLabels[0]; sink[i]<<=bL; sink[i]|=NONE; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
			} else { //left-left
				for(int l=0;l<leftLabels.length;++l){
					for(int l2=l+1;l2<leftLabels.length;++l2){ 
						sink[i]=t;   sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bL; sink[i]|=leftLabels[l]; sink[i]<<=bL; sink[i]|=leftLabels[l2]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
						sink[i]=t+1; sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bF; sink[i]|=tForm; sink[i]<<=bL; sink[i]|=leftLabels[l]; sink[i]<<=bL; sink[i]|=leftLabels[l2]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
					}
				}
			}
			t+=2;
			if(rightLabels.length==1){ //right-right
				sink[i]=t;   sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bL; sink[i]|=rightLabels[0]; sink[i]<<=bL; sink[i]|=NONE; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
				sink[i]=t+1; sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bF; sink[i]|=tForm; sink[i]<<=bL; sink[i]|=rightLabels[0]; sink[i]<<=bL; sink[i]|=NONE; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
			} else { //right-right
				for(int r=0;r<rightLabels.length;++r){
					for(int r2=r+1;r2<rightLabels.length;++r2){
						sink[i]=t;   sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bL; sink[i]|=rightLabels[r]; sink[i]<<=bL; sink[i]|=rightLabels[r2]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
						sink[i]=t+1; sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bF; sink[i]|=tForm; sink[i]<<=bL; sink[i]|=rightLabels[r]; sink[i]<<=bL; sink[i]|=rightLabels[r2]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
					}
				}
			}
			t+=2;
			//left-right
			for(int l=0;l<leftLabels.length;++l){
				for(int r=0;r<rightLabels.length;++r){
					sink[i]=t;   sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bL; sink[i]|=leftLabels[l]; sink[i]<<=bL; sink[i]|=rightLabels[r]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
					sink[i]=t+1; sink[i]<<=bT; sink[i]|=tPos; sink[i]<<=bD; sink[i]|=topMostVerbDepth; sink[i]<<=bF; sink[i]|=tForm; sink[i]<<=bL; sink[i]|=leftLabels[l]; sink[i]<<=bL; sink[i]|=rightLabels[r]; sink[i]<<=bD; sink[i]|=distToLastSB; ++i;
				}
			}
			t+=2;
		}

		final int s0l2f,s0l2p,s0l2l;
		final int s0r2f,s0r2p,s0r2l;
		final int s1l2f,s1l2p,s1l2l;
		final int s1r2f,s1r2p,s1r2l;


		if(s.s0L2>0){
			s0l2f=inst.form[s.s0L2];
			s0l2p=inst.tag[s.s0L2];
			s0l2l=s.getLabel(s.s0L2);
		} else {
			s0l2f=NONE;
			s0l2p=NONE;
			s0l2l=NONE;
		}
		if(s.s0R2>0){
			s0r2f=inst.form[s.s0R2];
			s0r2p=inst.tag[s.s0R2];
			s0r2l=s.getLabel(s.s0R2);
		} else {
			s0r2f=NONE;
			s0r2p=NONE;
			s0r2l=NONE;
		}
		if(s.tail.s0L2>0){
			s1l2f=inst.form[s.tail.s0L2];
			s1l2p=inst.tag[s.tail.s0L2];
			s1l2l=s.getLabel(s.tail.s0L2);
		} else {
			s1l2f=NONE;
			s1l2p=NONE;
			s1l2l=NONE;
		}
		if(s.tail.s0R2>0){
			s1r2f=inst.form[s.tail.s0R2];
			s1r2p=inst.tag[s.tail.s0R2];
			s1r2l=s.getLabel(s.tail.s0R2);
		} else {
			s1r2f=NONE;
			s1r2p=NONE;
			s1r2l=NONE;			
		}

		{
			//Add morphology combos on s0 and s1[lr], or s1 and s0[lr], but we ignore it if there is no dep there (that nothing is there should be made clear from other features anyway, e.g., tags on s[01][lr]
			{
				if(s0lp!=NONE){
					for(int a=0;a<s1m.length;++a){
						for(int b=0;b<s0lm.length;++b){
							sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0lp; //tags
							sink[i]<<=bM; sink[i]|=s1m[a]; sink[i]<<=bM; sink[i]|=s0lm[b];
							++i;
						}
					}
				}
				++t;
				if(s0rp!=NONE){
					for(int a=0;a<s1m.length;++a){
						for(int b=0;b<s0rm.length;++b){
							sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0rp; //tags
							sink[i]<<=bM; sink[i]|=s1m[a]; sink[i]<<=bM; sink[i]|=s0rm[b];
							++i;
						}
					}
				}
				++t;
				if(s1lp!=NONE){
					for(int a=0;a<s0m.length;++a){
						for(int b=0;b<s1lm.length;++b){
							sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1lp; //tags
							sink[i]<<=bM; sink[i]|=s0m[a]; sink[i]<<=bM; sink[i]|=s1lm[b];
						}
					}
				}
				++t;
				if(s1rp!=NONE){
					for(int a=0;a<s0m.length;++a){
						for(int b=0;b<s1rm.length;++b){
							sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1rp; //tags
							sink[i]<<=bM; sink[i]|=s0m[a]; sink[i]<<=bM; sink[i]|=s1rm[b];
						}
					}
				}
				++t;
			}
		}

		{ //Zhang and Nivre distance features

			int distBucketed=getIntBucket(Math.abs(s0-s1));
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s1f; ++t; ++i; //s1w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s1t; ++t; ++i; //s1t d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s0f; ++t; ++i; //s0w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s0p d

			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bF;sink[i]|=s1f;sink[i]<<=bF;sink[i]|=s0f; ++t; ++i; //s1w s0w d
			sink[i]=t; sink[i]<<=bD;sink[i]|=distBucketed;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s0t; ++t; ++i; //s1t s0t d
		}
		{ //Zhang and Nivre valency features
			int s0vL=getValencyBucket(s.getLeftDeps(s.s0));
			int s0vR=getValencyBucket(s.getRightDeps(s.s0));
			int s1vR,s1vL;
			if(s.tail==null){
				s1vR=s1vL=AbstractSymbolMapping.NONE_IDX;
			} else {
				s1vL=getValencyBucket(s.getLeftDeps(s.tail.s0));
				s1vR=getValencyBucket(s.getRightDeps(s.tail.s0));
			}
			sink[i]=t; sink[i]<<=bD; sink[i]|=s1vR; sink[i]<<=bF; sink[i]|=s1f; ++t; ++i; //s1wvR
			sink[i]=t; sink[i]<<=bD; sink[i]|=s1vL; sink[i]<<=bF; sink[i]|=s1f; ++t; ++i; //s1wvL
			sink[i]=t; sink[i]<<=bD; sink[i]|=s1vR; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i; //s1tvR
			sink[i]=t; sink[i]<<=bD; sink[i]|=s1vL; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i; //s1tvL

			sink[i]=t; sink[i]<<=bD; sink[i]|=s0vR; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i; //s0wvR //This is not in Zhang and Nivre
			sink[i]=t; sink[i]<<=bD; sink[i]|=s0vL; sink[i]<<=bF; sink[i]|=s0f; ++t; ++i; //s0wvL
			sink[i]=t; sink[i]<<=bD; sink[i]|=s0vR; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0tvR //This is not in Zhang and Nivre
			sink[i]=t; sink[i]<<=bD; sink[i]|=s0vL; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i; //s0tvL

		}
		{ //Zhang and Nivre unigrams
			//Useless in ArcStandard since nothing on the stack can have a head
			//			sink[i]=t;sink[i]<<=bF;sink[i]|=s1hf; ++t; ++i; //s0hw
			//			sink[i]=t;sink[i]<<=bT;sink[i]|=s1ht; ++t; ++i; //s0ht
			//			sink[i]=t;sink[i]<<=bL;sink[i]|=s1l;  ++t; ++i; //s0l  //label

			sink[i]=t;sink[i]<<=bF;sink[i]|=s1lf; ++t; ++i; //s1lw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1lp; ++t; ++i; //s1lt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1ll; ++t; ++i; //s1ll //label on leftmost

			sink[i]=t;sink[i]<<=bF;sink[i]|=s1rf; ++t; ++i; //s1rw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1rp; ++t; ++i; //s1rt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1rl; ++t; ++i; //s1rl //label on rightmost

			sink[i]=t;sink[i]<<=bF;sink[i]|=s0lf; ++t; ++i; //s0lw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0lp; ++t; ++i; //s0lt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s0ll; ++t; ++i; //s0ll //label on leftmost

			//This is mirrored from above, not in Zhang and Nivre
			sink[i]=t;sink[i]<<=bF;sink[i]|=s0rf; ++t; ++i; //s0rw
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0rp; ++t; ++i; //s0rt
			sink[i]=t;sink[i]<<=bL;sink[i]|=s0rl; ++t; ++i; //s0rl //label on rightmost

		}

		{ //Zhang and Nivre third-order
			//			S0h2 w; S0h2 p; S0h2 l; //can't do in arcstandard 
			//			S0l2 w; S0l2 p; S0l2 l;
			//			S0r2 w; S0r2 p; S0r2 l; 
			//			N0l2 w; N0l2 p; N0l2 l;
			//			S0 pS0l pS0l2 p; S0 pS0r pS0r2 p;
			//			S0 pS0h pS0h2 p; //can't do in arcstandard 
			//			N0 pN0l pN0l2 p;

			//Left-most-left-most and right-most-right-most, word, tag, and label

			sink[i]=t;sink[i]<<=bF;sink[i]|=s0l2f; ++t; ++i; //s0l2w
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0l2p; ++t; ++i; //s0l2t
			sink[i]=t;sink[i]<<=bL;sink[i]|=s0l2l; ++t; ++i; //s0l2l

			//The following three are not in Zhang and Nivre (since buffer front cannot have right dependents in arc eager)
			sink[i]=t;sink[i]<<=bF;sink[i]|=s0r2f; ++t; ++i; //s0r2w
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0r2p; ++t; ++i; //s0r2t
			sink[i]=t;sink[i]<<=bL;sink[i]|=s0r2l; ++t; ++i; //s0r2l

			sink[i]=t;sink[i]<<=bF;sink[i]|=s1l2f; ++t; ++i; //s1l2f
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1l2p; ++t; ++i; //s1l2t
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1l2l; ++t; ++i; //s1l2l

			sink[i]=t;sink[i]<<=bF;sink[i]|=s1r2f; ++t; ++i; //s1r2f
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1r2p; ++t; ++i; //s1r2t
			sink[i]=t;sink[i]<<=bL;sink[i]|=s1r2l; ++t; ++i; //s1r2l

			//Then the trigrams of s0, s0l, s0l2 tags. on s0 and s1, on both left and right
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0lp;sink[i]<<=bT;sink[i]|=s0l2p; ++t; ++i; //s0t s0lt s0l2t
			//this one is not in Zhang and Nivre since they don't have right deps on this token in arc eager
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0rp;sink[i]<<=bT;sink[i]|=s0r2p; ++t; ++i; //s0t s0rt s0r2t
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1lp;sink[i]<<=bT;sink[i]|=s1l2p; ++t; ++i; //s1t s1lt s1l2t
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1rp;sink[i]<<=bT;sink[i]|=s1r2p; ++t; ++i; //s1t s1rt s1r2t


			//Then I also want to add leftmost dependents rightmost dependent, and vice versa
			//			//The 4 below are is F1
			//			if(Options.x5){ //didn't help when I tried this last time XXX here we can remove stuff from state if we don't use these features
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0lp;sink[i]<<=bT;sink[i]|=s0lrp; ++t; ++i;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0rp;sink[i]<<=bT;sink[i]|=s0rlp; ++t; ++i;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1lp;sink[i]<<=bT;sink[i]|=s1lrp; ++t; ++i;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1rp;sink[i]<<=bT;sink[i]|=s1rlp; ++t; ++i;
			//			}
		}
		//XXX commented out for now, doesn't seem to help (but try again)
//		if(s.leftLabels!=null){ //Zhang and Nivre label set //The test here is true if we do labeled parsing, false otherwise
//			//We don't flag empty sets here, since this is already signalled by the valency counts on these tokens.
//
//			//We do make special treatment of the root node if it is s1 though, since root can have any number of right dependents when we do doc parsing
//
//			short[] s1Right=(s1==0?NONE_SHORT_ARR:s.rightLabels[s1]);
//
//			for(short b:s.leftLabels[s0]){ sink[i]=t;sink[i]<<=bF;sink[i]|=s0f;sink[i]<<=bL;sink[i]|=b; ++i;	} ++t; //s0wsL
//			for(short b:s.leftLabels[s0]){ sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bL;sink[i]|=b; ++i;	} ++t; //s0tsL
//			//these two are not in Zhang and Nivre
//			for(short b:s.rightLabels[s0]){ sink[i]=t;sink[i]<<=bF;sink[i]|=s0f;sink[i]<<=bL;sink[i]|=b; ++i; } ++t; //s0wsR
//			for(short b:s.rightLabels[s0]){ sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bL;sink[i]|=b; ++i; } ++t; //s0tsR
//
//			if(s1>=0){
//				for(short b:s.leftLabels[s1]){ sink[i]=t;sink[i]<<=bF;sink[i]|=s1f;sink[i]<<=bL;sink[i]|=b; ++i;	} ++t; //s1wsL
//				for(short b:s.leftLabels[s1]){ sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bL;sink[i]|=b; ++i;	} ++t; //s1tsL
//
//				for(short b:s1Right){ sink[i]=t;sink[i]<<=bF;sink[i]|=s1f;sink[i]<<=bL;sink[i]|=b; ++i; } ++t; //s1wsR
//				for(short b:s1Right){ sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bL;sink[i]|=b; ++i; } ++t; //s1tsR
//			} else {
//				t+=4;
//			}
//		}
		int len=inst.form.length;
		//Add some context features on every token

		final int s0Plus1t,s0Plus2t,s0Minus1t,s0Minus2t,s1Plus1t,s1Plus2t,s1Minus1t,s1Minus2t;
		{ //Assign context tags
			if(s0+1<len){
				s0Plus1t=inst.tag[s0+1];
				if(s0+2<len){
					s0Plus2t=inst.tag[s0+2];
				} else {
					s0Plus2t=NONE;
				}
			} else {
				s0Plus1t=NONE;
				s0Plus2t=NONE;
			}

			if(s1+1<len){
				s1Plus1t=inst.tag[s1+1];
				if(s1+2<len){
					s1Plus2t=inst.tag[s1+2];
				} else {
					s1Plus2t=NONE;
				}
			} else {
				s1Plus1t=NONE;
				s1Plus2t=NONE;
			}

			if(s0-1>=0){
				s0Minus1t=inst.tag[s0-1];
				if(s0-2>=0){
					s0Minus2t=inst.tag[s0-2];
				} else {
					s0Minus2t=NONE;
				}
			} else {
				s0Minus1t=NONE;
				s0Minus2t=NONE;
			}

			if(s1-1>=0){
				s1Minus1t=inst.tag[s1-1];
				if(s1-2>=0){
					s1Minus2t=inst.tag[s1-2];
				} else {
					s1Minus2t=NONE;
				}
			} else {
				s1Minus1t=NONE;
				s1Minus2t=NONE;
			}
		} //Done assigning context tags
		{ //Context tag features
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0Plus1t;sink[i]<<=bT;sink[i]|=s0Minus1t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0Plus1t;sink[i]<<=bT;sink[i]|=s0Plus2t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0Minus1t;sink[i]<<=bT;sink[i]|=s0Minus2t; ++i; ++t;

			sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1Plus1t;sink[i]<<=bT;sink[i]|=s1Minus1t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1Plus1t;sink[i]<<=bT;sink[i]|=s1Plus2t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1Minus1t;sink[i]<<=bT;sink[i]|=s1Minus2t; ++i; ++t;

			//			int s0Plus2w=(s0+2<len?inst.form[s0+2]:NONE);
			//			int s0Plus1w=(s0+1<len?inst.form[s0+1]:NONE);
			//			int s0Minus1w=(s0-1>=0?inst.form[s0-1]:NONE);
			//			int s0Minus2w=(s0-2>=0?inst.form[s0-2]:NONE);			
			//			int s1Plus1w=(s1+1<len?inst.form[s1+1]:NONE);
			//			int s1Plus2w=(s1+2<len?inst.form[s1+2]:NONE);
			//			int s1Minus1w=(s1-1>=0?inst.form[s1-1]:NONE);
			//			int s1Minus2w=(s1-2>=0?inst.form[s1-2]:NONE);


			//			if(false){ //Other combos with these two
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Minus1t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Minus2t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Plus1t;  ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Plus2t;  ++i; ++t;
			//
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Minus1t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Minus2t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Plus1t;  ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Plus2t;  ++i; ++t;
			//
			//			}

			//			if(Options.x1){ //didn't help when tried
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bF;sink[i]|=s0Plus1w;sink[i]<<=bF;sink[i]|=s0Minus1w; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bF;sink[i]|=s0Plus1w;sink[i]<<=bF;sink[i]|=s0Plus2w; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bF;sink[i]|=s0Minus1w;sink[i]<<=bF;sink[i]|=s0Minus2w; ++i; ++t;
			//				
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bF;sink[i]|=s1Plus1w;sink[i]<<=bF;sink[i]|=s1Minus1w; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bF;sink[i]|=s1Plus1w;sink[i]<<=bF;sink[i]|=s1Plus2w; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bF;sink[i]|=s1Minus1w;sink[i]<<=bF;sink[i]|=s1Minus2w; ++i; ++t;				
			//			}
			//			if(false){ //One step more context
			//				int s0Plus3t=(s0+3<len?inst.tag[s0+3]:NONE);
			//				int s0Minus3t=(s0-3>=0?inst.tag[s0-3]:NONE);
			//				int s1Plus3t=(s1+3<len?inst.tag[s1+3]:NONE);
			//				int s1Minus3t=(s1-3>=0?inst.tag[s1-3]:NONE);
			//
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0Plus1t;sink[i]<<=bT;sink[i]|=s0Plus2t;  sink[i]<<=bT; sink[i]|=s0Plus3t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s0Minus1t;sink[i]<<=bT;sink[i]|=s0Minus2t;sink[i]<<=bT; sink[i]|=s0Minus3t;++i; ++t;
			//
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1Plus1t;sink[i]<<=bT;sink[i]|=s1Plus2t;  sink[i]<<=bT; sink[i]|=s1Plus3t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=s1Minus1t;sink[i]<<=bT;sink[i]|=s1Minus2t;sink[i]<<=bT; sink[i]|=s1Minus3t;++i; ++t;
			//				
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Minus3t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Plus3t;  ++i; ++t;
			//
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Minus3t; ++i; ++t;
			//				sink[i]=t;sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Plus3t;  ++i; ++t;
			//				
			//			}
		}
		//		if(Options.x2){ //didn't help when tried
		//			//Add between tags
		//			for(int b=Math.min(s0, s1)+1,m=Math.max(s0, s1);b<m;++b){
		//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=inst.tag[b]; ++i;
		//			}
		//			++t;
		//		}

		//		if(Options.x3){ //didn't help when tried
		//			//Add between forms
		//			for(int b=Math.min(s0, s1)+1,m=Math.max(s0, s1);b<m;++b){
		//				sink[i]=t;sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bF;sink[i]|=inst.form[b]; ++i;
		//			}
		//			++t;
		//		}

		//		{ //Added
		//		//Added Singles: deeper in buffer and stack
		//		sink[i]=t; sink[i]<<=bF; sink[i]|=s2f; ++t; ++i;                             //s2w
		//		sink[i]=t; sink[i]<<=bT; sink[i]|=s2t; ++t; ++i;                             //s2t
		//		sink[i]=t; sink[i]<<=bF; sink[i]|=s2f; sink[i]<<=bT; sink[i]|=s2t; ++t; ++i; //s2wt
		//		
		//		sink[i]=t; sink[i]<<=bF; sink[i]|=b2f; ++t; ++i;                             //b2w
		//		sink[i]=t; sink[i]<<=bT; sink[i]|=b2t; ++t; ++i;                             //b2t
		//		sink[i]=t; sink[i]<<=bF; sink[i]|=b2f; sink[i]<<=bT; sink[i]|=b2t; ++t; ++i; //b2wt
		//
		//Added pairs of s0+b0
		{
			sink[i]=t; sink[i]<<=bF;sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bF; sink[i]|=b0f; ++t; ++i; //s1wt b0w
			sink[i]=t; sink[i]<<=bF;sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i; //s1wt b0t
			sink[i]=t; sink[i]<<=bF;sink[i]|=s0f; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i; //s1w b0wt
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i; //s1t b0wt
		}

		{ //s1+b0
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bF; sink[i]|=b0f; ++t; ++i;
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;
			sink[i]=t; sink[i]<<=bF;sink[i]|=s1f; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;
			sink[i]=t; sink[i]<<=bT;sink[i]|=s1t; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bT; sink[i]|=b0t; ++t; ++i;
		}

		//		//XXX try s2+s0
		//		if(false){//s2+s1
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s2f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=s1f; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s2f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s2f; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bT;sink[i]|=s2t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;
		//		}
		////	}
		//		
		//		if(false){
		//			int s0Lemma=inst.lemma[s0];
		//			int s1Lemma=inst.lemma[s1];
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s0Lemma; sink[i]<<=bF; sink[i]|=s1Lemma; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s0Lemma; sink[i]<<=bF; sink[i]|=s1Lemma; sink[i]<<=bT; sink[i]|=s0t; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s0Lemma; sink[i]<<=bF; sink[i]|=s1Lemma; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;
		//			sink[i]=t; sink[i]<<=bF;sink[i]|=s0Lemma; sink[i]<<=bF; sink[i]|=s1Lemma; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++t; ++i;
		//		}

		{ //More lookahead in buffer (b2)
			sink[i]=t; sink[i]<<=bF; sink[i]|=b2f; ++t; ++i;                             //b2w
			sink[i]=t; sink[i]<<=bT; sink[i]|=b2t; ++t; ++i;                             //b2t
			sink[i]=t; sink[i]<<=bF; sink[i]|=b2f; sink[i]<<=bT; sink[i]|=b2t; ++t; ++i; //b2wt
			sink[i]=t; sink[i]<<=bT;sink[i]|=s0t;sink[i]<<=bT;sink[i]|=s1t;sink[i]<<=bT;sink[i]|=b2t; ++t; ++i;	 //s0t s1t b2t
		}

		//		if(false){ //More lookahead in buffer (b3)
		//			int b3=s.buf.get(3);
		//			final int b3t,b3f;
		//			if(b3>0){
		//				b3t=inst.tag[b3];
		//				b3f=inst.form[b3];
		//			} else {
		//				b3t=NONE;
		//				b3f=NONE;
		//			}
		//			
		//			sink[i]=t; sink[i]<<=bF; sink[i]|=b3f; ++t; ++i;                             //b3w
		//			sink[i]=t; sink[i]<<=bT; sink[i]|=b3t; ++t; ++i;                             //b3t
		//			sink[i]=t; sink[i]<<=bF; sink[i]|=b3f; sink[i]<<=bT; sink[i]|=b3t; ++t; ++i; //b3wt
		//			sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT;sink[i]|=b3t; ++t; ++i;	 //s0t s1t b3t					
		//
		//		}
		//		
		//		if(false){ //Bagged lookahead (5 tokens, i.e., b2,b3,b4,b5,b6)
		//			for(int b=2;i<7;++i){
		//				int bq=s.buf.get(b);
		//				if(bq==-1)
		//					break;
		//				int bqt=inst.tag[bq];
		//				int bqf=inst.form[bq];
		//
		//				sink[i]=t;   sink[i]<<=bF; sink[i]|=bqf; ++i;                             //bqw
		//				sink[i]=t+1; sink[i]<<=bT; sink[i]|=bqt; ++i;                             //bqt
		//				sink[i]=t+2; sink[i]<<=bF; sink[i]|=bqf; sink[i]<<=bT; sink[i]|=bqt; ++i; //bqwt
		//				sink[i]=t+3; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT;sink[i]|=bqt; ++i;	 //s0t s1t bqt
		//			}
		//			t+=4;
		//		}
		//		


		{ //More lookadeep in stack (s2)
			sink[i]=t; sink[i]<<=bF; sink[i]|=s2f; ++t; ++i;                             //s2w
			sink[i]=t; sink[i]<<=bT; sink[i]|=s2t; ++t; ++i;                             //s2t
			sink[i]=t; sink[i]<<=bF; sink[i]|=s2f; sink[i]<<=bT; sink[i]|=s2t; ++t; ++i; //s2wt
			sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT;sink[i]|=s2t; ++t; ++i;	 //s0t s1t s2t
		}
		//		
		//		if(false){ //More lookadeep in stack (s3)
		//			final int s3;
		//			if(s.tail.tail==null || s.tail.tail.tail==null)
		//				s3=-1;
		//			else
		//				s3=s.tail.tail.tail.s0;
		//			
		//			final int s3t,s3f;
		//			if(s3>-1){
		//				s3t=inst.tag[s3];
		//				s3f=inst.form[s3];
		//			} else {
		//				s3t=NONE;
		//				s3f=NONE;					
		//			}			
		//			sink[i]=t; sink[i]<<=bF; sink[i]|=s3f; ++t; ++i;                             //s3w
		//			sink[i]=t; sink[i]<<=bT; sink[i]|=s3t; ++t; ++i;                             //s3t
		//			sink[i]=t; sink[i]<<=bF; sink[i]|=s3f; sink[i]<<=bT; sink[i]|=s3t; ++t; ++i; //s3wt
		//			sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT;sink[i]|=s3t; ++t; ++i;	 //s0t s1t s3t
		//		}

		//		if(false){ //Bagged lookahead (5 tokens, i.e., s2,s3,s4,s5,s6)
		//			State q=s.tail.tail;
		//			for(int d=0;q!=null && d<5;++d){
		//				int sq=q.s0;
		//				int sqt=inst.tag[sq];
		//				int sqf=inst.form[sq];
		//
		//				sink[i]=t;   sink[i]<<=bF; sink[i]|=sqf; ++i;                             //sqw
		//				sink[i]=t+1; sink[i]<<=bT; sink[i]|=sqt; ++i;                             //sqt
		//				sink[i]=t+2; sink[i]<<=bF; sink[i]|=sqf; sink[i]<<=bT; sink[i]|=sqt; ++i; //sqwt
		//				sink[i]=t+3; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT;sink[i]|=sqt; ++i;	 //s0t s1t sqt
		//				
		//				q=q.tail;
		//			}
		//			t+=4;
		//		}

		{ //is b0 swapped back onto the buffer?
			int b0Swapped=(s.buf.hasSwappedItems()?1:0);
			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; ++i; ++t;
			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
		}
		//
		//		if(Options.x1){
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
		//			
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bF; sink[i]|=s1f; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bF; sink[i]|=s1f; ++i; ++t;
		//		}
		//
		//		if(Options.x2){
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=b0t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=b0f; ++i; ++t;
		//		}
		//		
		//		if(Options.x3){
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=b0t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=b0f; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=b1t; ++i; ++t;
		//			sink[i]=t; sink[i]<<=1; sink[i]|=b0Swapped; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=b1f; ++i; ++t;
		//
		//		}

		//XXX: to try:
		//Morphology on context
		//Morphology in between
		//Also, try making larger/smaller context for tags

		//Also, try features inspired by the Manning 2014 EMNLP neural network parser paper (specifically the second leftmost and second rightmost deps)

		//No this doesn't seem to help...

		//		{ // heads on context
		//			final int s0p1lbl=(s0+1<len?s.getLabel(s0+1):NONE);
		//			final int s0m1lbl=(s0-1>=0?s.getLabel(s0-1):NONE);
		//			final int s1p1lbl=(s1+1<len?s.getLabel(s1+1):NONE);
		//			final int s1m1lbl=(s1-1>=0?s.getLabel(s1-1):NONE);
		//
		//			if(Options.x7){ //TODO RERUN; was x1
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1m1lbl; ++i; ++t;
		//			}
		//			
		//			if(Options.x8){
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1m1lbl; ++i; ++t;
		//			}
		//	
		//			if(Options.x2){
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s0Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s0Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s1Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//			}
		//
		//			if(Options.x3){ //here?
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=s0Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//			}
		//
		//			if(Options.x4){
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bL; sink[i]|=s1m1lbl; ++i; ++t;
		//			}
		//	
		//			if(Options.x5){
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s0Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s1Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//			}
		//
		//			if(Options.x6){
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s0Plus1t;  sink[i]<<=bL; sink[i]|=s0p1lbl; ++i; ++t;
		//				sink[i]=t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bT; sink[i]|=s0Minus1t; sink[i]<<=bL; sink[i]|=s0m1lbl; ++i; ++t;
		//			}
		//		}


//		i = extractStagStackFeatures(sink, inst, t, i, s0, s0t, s1, s1t, s0lp, s0rp, s1lp, s1rp);


		return i;
	}

	private int addDistBucketSBFeatures(int distToLastSB, int i, int t,int s0t, int s1t,int s0f, int s1f,int b0t, int b0f,int b0shape,int b0m1shape,int b0fcs,int stackSize) {
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		//Same as above, but with some forms
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bF; sink[i]|=s0f; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bF; sink[i]|=s1f; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bF; sink[i]|=b0f; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		if(false){ //changing this to true will add +2 to t
//			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bT; sink[i]|=topMostVerbPos; ++t; ++i;
//			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; sink[i]<<=bF; sink[i]|=topMostVerbForm; ++t; ++i;
		}
		
		if(useShapes){
			//And some shapes
			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bSH; sink[i]|=b0shape; sink[i]<<=bSH; sink[i]|=b0m1shape; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
			sink[i]=t; sink[i]<<=bD; sink[i]|=distToLastSB; sink[i]<<=bCS; sink[i]|=b0fcs; sink[i]<<=bT; sink[i]|=b0t; sink[i]<<=bD; sink[i]|=stackSize; ++t; ++i;
		}
		return i;
	}

	private int extractB0Context(Instance inst, int b0, int b0t, int t, int i) {
		//XXX NOTE: This function should ideally return t and i. We only return i. So the caller has to increase t by what t is increased by in here!
		if(useB0Context){
			final int b0Plus1t,b0Plus2t,b0Minus1t,b0Minus2t;
			final int b0Plus1f,b0Plus2f,b0Minus1f,b0Minus2f;
			if(b0+1<inst.form.length){
				b0Plus1t=inst.tag[b0+1];
				b0Plus1f=inst.form[b0+1];
				if(b0+2<inst.form.length){
					b0Plus2t=inst.tag[b0+2];
					b0Plus2f=inst.form[b0+2];				
				} else {
					b0Plus2t=NONE;
					b0Plus2f=NONE;
				}
			} else {
				b0Plus1t=NONE;
				b0Plus2t=NONE;
				b0Plus1f=NONE;
				b0Plus2f=NONE;
			}
			if(b0-1>=0){
				b0Minus1t=inst.tag[b0-1];
				b0Minus1f=inst.form[b0-1];
				if(b0-2>=0){
					b0Minus2t=inst.tag[b0-2];
					b0Minus2f=inst.form[b0-2];				
				} else {
					b0Minus2t=NONE;
					b0Minus2f=NONE;
				}
			} else {
				b0Minus1t=NONE;
				b0Minus2t=NONE;
				b0Minus1f=NONE;
				b0Minus2f=NONE;
			}
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bT;sink[i]|=b0Plus1t;sink[i]<<=bT;sink[i]|=b0Minus1t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bT;sink[i]|=b0Plus1t;sink[i]<<=bT;sink[i]|=b0Plus2t; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bT;sink[i]|=b0Minus1t;sink[i]<<=bT;sink[i]|=b0Minus2t; ++i; ++t;

			sink[i]=t;sink[i]<<=bF;sink[i]|=b0Plus1f; ++i; ++t;
			sink[i]=t;sink[i]<<=bF;sink[i]|=b0Plus2f; ++i; ++t;
			sink[i]=t;sink[i]<<=bF;sink[i]|=b0Minus1f; ++i; ++t;
			sink[i]=t;sink[i]<<=bF;sink[i]|=b0Minus2f; ++i; ++t;
			
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bF;sink[i]|=b0Plus1f; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bF;sink[i]|=b0Plus2f; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bF;sink[i]|=b0Minus1f; ++i; ++t;
			sink[i]=t;sink[i]<<=bT;sink[i]|=b0t;sink[i]<<=bF;sink[i]|=b0Minus2f; ++i; ++t;
			
			//There are 11 ++t until this point
		}
		return i;
	}

//	private int extractStagStackFeatures(long[] sink, Instance inst, int t,
//			int i, final int s0, final int s0t, final int s1, final int s1t,
//			final int s0lp, final int s0rp, final int s1lp, final int s1rp) {
//		if(useStag){ //There's a bug somewhere in here due to copy and pasting. Check the table on paper.
//			final int s0Stag=inst.stag[s0];
//			final int s0HeadDir=inst.stagHeadDir[s0];
//			final int s0WantsLeftDep=inst.stagDeps[s0][Instance.STAG_LEFT];
//			final int s0WantsRightDep=inst.stagDeps[s0][Instance.STAG_RIGHT];
//			final int s0StagLbl=inst.stagLbl[s0];
//
//			final int s1Stag=inst.stag[s1];
//			final int s1HeadDir=inst.stagHeadDir[s1];
//			final int s1WantsLeftDep=inst.stagDeps[s1][Instance.STAG_LEFT];
//			final int s1WantsRightDep=inst.stagDeps[s1][Instance.STAG_RIGHT];
//			final int s1StagLbl=inst.stagLbl[s1];
//
//			//Full tag
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s0Stag; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s0Stag;                                        sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s0Stag; sink[i]<<=bT; sink[i]|=s0lp;           sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s0Stag; sink[i]<<=bT; sink[i]|=s0rp;           sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			//Take it apart
//			sink[i]=t;sink[i]<<=3;sink[i]|=s0WantsLeftDep;                                 sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s1HeadDir;                                      sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s0WantsLeftDep;sink[i]<<=3;sink[i]|=s1HeadDir;  sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bL;sink[i]|=s0StagLbl;                                     sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s0WantsLeftDep; sink[i]<<=bT; sink[i]|=s0lp;    sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s0WantsRightDep; sink[i]<<=bT; sink[i]|=s0rp;   sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//
//			//Full tag
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s1Stag; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s1Stag;                                        sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s1Stag; sink[i]<<=bT; sink[i]|=s1lp;           sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bS;sink[i]|=s1Stag; sink[i]<<=bT; sink[i]|=s1rp;           sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			//Take it apart
//			sink[i]=t;sink[i]<<=3;sink[i]|=s1WantsRightDep;                                sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s0HeadDir;                                      sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s1WantsRightDep;sink[i]<<=3;sink[i]|=s0HeadDir; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bL;sink[i]|=s1StagLbl;                                     sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s1WantsLeftDep; sink[i]<<=bT; sink[i]|=s1lp;    sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=3;sink[i]|=s1WantsRightDep; sink[i]<<=bT; sink[i]|=s1rp;   sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//		}
//
//		if(useStack){
//			final int s0HeadOfS1,s1HeadOfS0,headOnLeftS0,s0StackLbl,headOnLeftS1,s1StackLbl,s0NeedsLeftDeps,s0NeedsRightDeps,s1NeedsLeftDeps,s1NeedsRightDeps;
//			s0StackLbl=inst.stackLbl[s0];
//			s0NeedsLeftDeps=inst.stackHasLeftDep[s0]?1:0;
//			s0NeedsRightDeps=inst.stackHasRightDep[s0]?1:0;
//			s0HeadOfS1=inst.stackHead[s1]==s0?1:0;
//			s1HeadOfS0=inst.stackHead[s0]==s1?1:0;
//			headOnLeftS0=inst.stackHead[s0]<s0?1:0;
//			headOnLeftS1=inst.stackHead[s1]<s1?1:0;
//			s1StackLbl=inst.stackLbl[s1];
//			s1NeedsLeftDeps=inst.stackHasLeftDep[s1]?1:0;
//			s1NeedsRightDeps=inst.stackHasRightDep[s1]?1:0;
//
//			sink[i]=t;sink[i]<<=2;sink[i]|=s0HeadOfS1;   sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=s1HeadOfS0;   sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=headOnLeftS0; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=headOnLeftS1; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bL; sink[i]|=s0StackLbl; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=bL; sink[i]|=s1StackLbl; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//
//			sink[i]=t;sink[i]<<=2;sink[i]|=s0NeedsLeftDeps;  sink[i]<<=bT; sink[i]|=s0lp; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=s0NeedsRightDeps; sink[i]<<=bT; sink[i]|=s0rp; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=s1NeedsLeftDeps;  sink[i]<<=bT; sink[i]|=s1lp; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//			sink[i]=t;sink[i]<<=2;sink[i]|=s1NeedsRightDeps; sink[i]<<=bT; sink[i]|=s1rp; sink[i]<<=bT; sink[i]|=s0t; sink[i]<<=bT; sink[i]|=s1t; ++i; ++t;
//		}
//		return i;
//	}


}
