package ims.trans.labelfilter;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.IntPair;
import ims.util.MutableInt;

public class PosBasedLabelFilter implements ILabelFilter {
	private static final long serialVersionUID = -1697977668410159330L;
	
	//We index this [headPos][depPos]
	final int[][][] labels;
	final int[] defaultLabel;
	
	public PosBasedLabelFilter(SymbolTable symTab,List<Instance> instances){
		Map<IntPair,Set<Integer>> seen=new HashMap<IntPair,Set<Integer>>();
		Map<Integer,MutableInt> lblCounts=new TreeMap<Integer,MutableInt>(); //To track most frequent
		int maxPos=0;
		for(Instance is:instances){
			for(int dep=1;dep<is.tag.length;++dep){
				int hPos=is.tag[is.head[dep]];
				int dPos=is.tag[dep];
				int lbl=is.lbl[dep];
				MutableInt.incrementMapValue(lblCounts, lbl);
				addLabel(seen,hPos,dPos,lbl);
				maxPos=Math.max(maxPos, Math.max(hPos, dPos));
			}
		}
		this.labels=new int[maxPos+1][maxPos+1][];
		for(Entry<IntPair,Set<Integer>> e:seen.entrySet()){
			IntPair ip=e.getKey();
			int hPos=ip.i1;
			int dPos=ip.i2;
			Set<Integer> s=e.getValue();
			this.labels[hPos][dPos]=new int[s.size()];
			int q=0;
			for(Integer i:s)
				this.labels[hPos][dPos][q++]=i.intValue();
		}
		this.defaultLabel=new int[]{MutableInt.getEntriesSortedByFrequencyDescending(lblCounts).get(0).getKey().intValue()};
	}

	private void addLabel(Map<IntPair, Set<Integer>> seen,int hPos,int dPos,int lbl) {
		IntPair ip=IntPair.intPairPool.get(hPos, dPos);
		Set<Integer> s=seen.get(ip);
		if(s==null){
			s=new TreeSet<Integer>(); //ensures that stuff is ordered and stays deterministic
			seen.put(ip, s);
		}
		s.add(lbl);
	}

	@Override
	public int[] getAllowedLabels(int head, int dep, Instance is) {
		int hPos=is.tag[head];
		int dPos=is.tag[dep];
		int[] r=labels[hPos][dPos];
		if(r==null)
			return defaultLabel;
		else
			return r;
	}
	
}
