package ims.trans.transitions;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.State;

public class DefaultArcStandardTransitionFilter implements ITransitionFilter {
	private static final long serialVersionUID = 6033436057574854017L;
	
	final int[] SH;
	final int[] SH_RA_LA;
	final int[] SH_RA;
	final int[] RA;
	final int[] RA_LA;
	
	public DefaultArcStandardTransitionFilter(SymbolTable symTab){
		int leftArcs=0;
		int rightArcs=0;
		
		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition tr=symTab.transition2label[i];
			switch(tr.getTR()){
			case RAS: ++rightArcs; break;
			case LA:  ++leftArcs;  break;
			default:  //do nothing
			}
		}
		
		
		SH=new int[1];                           
		SH_RA_LA=new int[1+leftArcs+rightArcs];  int b=0;
		SH_RA=new int[1+rightArcs];              int c=0;
		RA=new int[rightArcs];                   int d=0;
		RA_LA=new int[leftArcs+rightArcs];       int e=0;

		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition trans=symTab.transition2label[i];
			int id=trans.getId();
			TR tr=trans.getTR();
			SH_RA_LA[b++]=id;
			if(tr==TR.SH){
				SH[0]=id;
				SH_RA[c++]=id;
			} else if(tr==TR.LA){
				RA_LA[e++]=id;
			} else if(tr==TR.RAS){
				SH_RA[c++]=id;
				RA[d++]=id;
				RA_LA[e++]=id;
			}
		}
	}

	@Override
	public int[] getAllowedTransitions(State s, Instance inst) {
		if(s.tail==null)
			return SH;
		boolean canLeftArc=s.tail.s0!=0;
		boolean canShift=s.buf.size()>0;
		if(canLeftArc){
			if(canShift)
				return SH_RA_LA;
			else
				return RA_LA;
		} else {
			if(canShift)
				return SH_RA;
			else
				return RA;
		}
	}
	
	@Override
	public boolean isPOSFiltered(){
		return false;
	}
	
}
