package ims.trans.util;

import ims.data.ISentence;
import ims.io.Format;
import ims.io.StreamIteratorReader;
import ims.util.DisjointSetZeroToN;
import ims.util.DisjointSetZeroToNRankPathComp;
import ims.util.NonProjInfo;
import ims.util.Util;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;


public class ArcStandardPrecedence {

	static final boolean WITH_TRANSITIVITY=true; //XXX setting this to false doesn't work -- doesn't handle transitivity among siblings properly
	
	final List<Set<Integer>> p;
	public final boolean[][] anc;
	public final DisjointSetZeroToN projSubTrees;
//	public final boolean[] isNonProj;
	
	public ArcStandardPrecedence(int[] heads){
		this.anc=NonProjInfo.getAncestors(heads);
		this.p=getPrecedence(heads,anc);
//		this.isNonProj=NonProjInfo.getNonProjArcs(heads);
		this.projSubTrees=new DisjointSetZeroToNRankPathComp(heads.length);
//		for(int i=1;i<heads.length;++i){
//			if(!isNonProj[i])
//				projSubTrees.union(i, heads[i]);
//		}
	}
	
	//True if d1 must be attached before d2
	public boolean precedes(int d1,int d2){
		return p.get(d2).contains(d1);
	}
	
	
	//XXX This can be made alot more efficient by doing it bottom up, but this is a quick and dirty solution for now. Let's see if it turns out useful anyway
	public static List<Set<Integer>> getPrecedence(int[] heads,boolean[][] anc){
		List<Set<Integer>> r=new ArrayList<Set<Integer>>();
		r.add(new TreeSet<Integer>());
		Rep rep=new Rep(heads);
		for(int dep=1;dep<heads.length;++dep){
			r.add(getPrecedence(dep,heads,rep,anc));
		}
		return r;
	}
	
	public static Set<Integer> getPrecedence(int dep,int[] heads,Rep rep,boolean[][] anc){
		Set<Integer> before=new TreeSet<Integer>();
		
		//There are two issues:
		//-- descendants must come before
		//-- siblings that are closer to the head must come before
		
		//And by transitivity:
		//-- descendants of siblings must come first
		
		//This array is index [anc][desc]
		int head=heads[dep];
		boolean myHeadOnRight=head>dep;
		if(myHeadOnRight){
			List<Integer> leftHeadDeps=rep.leftDeps.get(head);
			for(int i=leftHeadDeps.size()-1;i>=0;--i){
				int l=leftHeadDeps.get(i);
				if(l==dep)
					break;
				//Else l and all it's descendants need to be done first
				before.add(l); //Add l
				if(WITH_TRANSITIVITY){
					for(int q=1;q<heads.length;++q){ //Add l's descendants
						if(anc[l][q])
							before.add(q);
					}
				}
			}
		} else {
			List<Integer> rightHeadDeps=rep.rightDeps.get(head);
			for(int i=0;i<rightHeadDeps.size();++i){
				int r=rightHeadDeps.get(i);
				if(r==dep)
					break;
				before.add(r);
				if(WITH_TRANSITIVITY){
					for(int q=1;q<heads.length;++q){
						if(anc[r][q])
							before.add(q);
					}
				}
			}
		}
		//Then add all the descendants of dep itself
		if(WITH_TRANSITIVITY){
			for(int q=1;q<heads.length;++q)
				if(anc[dep][q])
					before.add(q);
		} else {
			for(int q=1;q<heads.length;++q)
				if(heads[q]==dep)
					before.add(q);
		}
		return before;
	}
	
	static class Rep {
		List<List<Integer>> leftDeps=new ArrayList<List<Integer>>();
		List<List<Integer>> rightDeps=new ArrayList<List<Integer>>();
		
		public Rep(int[] heads){
			for(int i=0;i<heads.length;++i){
				leftDeps.add(new ArrayList<Integer>());
				rightDeps.add(new ArrayList<Integer>());
			}
			for(int i=1;i<heads.length;++i){
				boolean left=heads[i]>i;
				if(left)
					leftDeps.get(heads[i]).add(i);
				else
					rightDeps.get(heads[i]).add(i);
			}
		}
	}

	public static void main(String[] args) throws UnsupportedEncodingException, IOException{
		int[][] pairs=new int[][]{{5,3},{2,5},{1,2},{2,1}};
		Iterator<ISentence> it=new StreamIteratorReader(System.in,Format.CoNLL06);
		while(it.hasNext()){
			ISentence s=it.next();
			int[] head=s.getGoldHeads();
			ArcStandardPrecedence asp=new ArcStandardPrecedence(head);
//			List<Set<Integer>> p=getPrecedence(s.head);
			for(int i=0;i<asp.p.size();++i){
				System.out.printf("%-3d   %s\n",i,Util.joinStrings(asp.p.get(i), ","));
			}
			System.out.println();
			for(int[] pair:pairs){
				int x=pair[0];
				int y=pair[1];
				System.out.printf("%d < %d ? %d\n",x,y,(asp.precedes(x, y)?1:0));
			}
			System.out.println();
			//Print projective subtrees
			StringBuilder sb=new StringBuilder();
			Map<Integer,Set<Integer>> m=new TreeMap<Integer,Set<Integer>>();
			for(int i=0;i<head.length;++i){
				int root=asp.projSubTrees.find(i);
				Set<Integer> set=m.get(root);
				if(set==null){
					set=new TreeSet<Integer>();
					m.put(root, set);
				}
				set.add(i);
			}
			for(Entry<Integer,Set<Integer>> e:m.entrySet()){
				sb.append(String.format("%-3d   %s\n",e.getKey(),Util.joinStrings(e.getValue(), ",")));
			}
			System.out.println(sb.toString());
		}
		
		//try some random pairs:
		
	}
}
