package ims.trans.util;

import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

import ims.data.CoNLL06Sentence;
import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.trans.Options;

public class TagBasedMajorityBaselineParser {

	static class MapEntry {
		int leftHead=0;
		int rightHead=0;
		public boolean attachLeft(){
			return leftHead>rightHead;
		}
	}
	
	private static void applyMajorityVote(SentenceReader<?> testReader,	SentenceWriter writer, Map<String, MapEntry> counts, boolean defaultLeft) throws IOException {
		for(ISentence s:testReader){
			CoNLL06Sentence cast=(CoNLL06Sentence) s;
			int[] pHeads=new int[s.getLength()];
			String[] tags=cast.getCPos();
			for(int i=1;i<tags.length-1;++i){
				boolean left=headOnLeft(counts,defaultLeft,tags[i]);
				if(left)
					pHeads[i]=i-1;
				else
					pHeads[i]=i+1;
			}
			pHeads[pHeads.length-1]=pHeads.length-2; //The final token always has it's head on the left
			s.setPredHeads(pHeads);
			writer.write(s);
		}
	}

	private static boolean headOnLeft(Map<String, MapEntry> counts, boolean defaultLeft, String tag) {
		MapEntry e=counts.get(tag);
		if(e==null){
			return defaultLeft;
		} else {
			return e.attachLeft();
		}
	}

	private static Map<String, MapEntry> doCounts(SentenceReader<?> trainReader) {
		Map<String,MapEntry> m=new TreeMap<String,MapEntry>();
		for(ISentence s:trainReader){
			CoNLL06Sentence cast=(CoNLL06Sentence) s;
			for(int i=1;i<s.getLength();++i){
				String depTag=cast.getCPos()[i];
				int depHead=s.getTrainHeads()[i];
				boolean left=depHead<i;
				addCount(m,depTag,left);
			}
		}
		return m;
	}

	private static void addCount(Map<String, MapEntry> m, String depTag, boolean left) {
		MapEntry e=m.get(depTag);
		if(e==null){
			e=new MapEntry();
			m.put(depTag, e);
		}
		if(left)
			e.leftHead++;
		else
			e.rightHead++;
	}

	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		SentenceReader<?> trainReader=ReaderWriterFactory.getTrainReader(options);
		Map<String,MapEntry> counts=doCounts(trainReader);
		
		SentenceReader<?> testReader=ReaderWriterFactory.getTestReader(options);
		SentenceWriter writer=ReaderWriterFactory.getWriter(options.out);
		
		applyMajorityVote(testReader,writer,counts,true);
		writer.close();
	}

}
