package ims.util;

import gnu.trove.iterator.TDoubleIterator;
import gnu.trove.list.TDoubleList;
import gnu.trove.list.array.TDoubleArrayList;

public abstract class DoubleSum {

	public abstract double getSum();
	protected abstract void addAddend(double d);
	
	protected TDoubleList addends=new TDoubleArrayList();

	public void addDouble(double d){
		addends.add(d);
		addAddend(d);
	}
	
	public double addendCount(){
		return addends.size();
	}
	
	public double getAvg(){
		return getSum()/addendCount();
	}

	public double getStd(){
		TDoubleIterator it=addends.iterator();
		DoubleSum stdSum;
		try {
			stdSum=this.getClass().newInstance();
		} catch (Exception e){
			throw new Error("Error here!");
		}
		
		double avg=getAvg();
		
		while(it.hasNext()){
			double d=it.next();
			d-=avg;
			stdSum.addDouble(d*d);
		}
		
//		double s=Math.sqrt(stdSum.getSum()/(addendCount()));
		double s=Math.sqrt(stdSum.getSum()/(addendCount()-1)); //Bessels correction
		
		return s;
	}
	

	public static class VanillaDoubleSum extends DoubleSum {

		private double sum=0.d;
		
		@Override
		public double getSum() {
			return sum;
		}

		@Override
		public void addAddend(double d) {
			sum+=d;
		}
		
	}
	
	public static class KahanDoubleSum extends DoubleSum {

		private double sum=0.d;
		private double c=0.d;
		
		@Override
		public double getSum() {
			return sum;
		}

		@Override
		public void addAddend(double d) {
			double y=d-c;
			double t=sum+y;
			c=(t-sum)-y;
			sum=t;
		}
		
	}

	
	
	public static DoubleSum getDoubleSum(boolean kahan){
		if(kahan)
			return new KahanDoubleSum();
		else
			return new VanillaDoubleSum();
	}
	
}
