package ims.util;

import java.io.Serializable;
import java.util.Random;

/*
 * sampleEntry and logSumExp is based on code from the RBGParser by
 * Tao Lei and Yuan Zhang (https://github.com/taolei87/RBGParser)
 */

public class ProbSampler implements Serializable {
	private static final long serialVersionUID = 2435683183402922664L;
	
	private final Random r;
	
	public ProbSampler(){
		this(System.currentTimeMillis());
	}
	
	public ProbSampler(long seed){
		r=new Random(seed);
	}
	
	public int sampleEntry2(double[] scores){
		double[] softMaxed=new double[scores.length];
		double expSum=0.d;
		for(int i=0;i<scores.length;++i){
			double e=Math.exp(scores[i]);
			softMaxed[i]=e;
			expSum+=e;
		}
		for(int i=0;i<scores.length;++i){
			softMaxed[i]/=expSum;
		}

		int idx=rouletteSampleEntry(r, softMaxed, 1.0d);
		return idx;
	}
	
	public int sampleEntry(double[] scores){
    	double sumScore = Double.NEGATIVE_INFINITY;
    	for (int i = 0; i < scores.length; i++) {
    		sumScore = logSumExp(sumScore, scores[i]);
    	}
    	double logp = Math.log(r.nextDouble() + 1e-60);
    	double cur = Double.NEGATIVE_INFINITY;
    	int ret = 0;
    	for (; ret < scores.length; ret++) {
    		cur = logSumExp(cur, scores[ret]);
    		if (logp + sumScore < cur)
    			break;
    	}
    	return ret;
	}

	//This function returns log(e^x+e^y)
	public static double logSumExp(double x, double y){
		if (x == Double.NEGATIVE_INFINITY && x == y)
			return Double.NEGATIVE_INFINITY;
		else if (x < y)
			return y + Math.log1p(Math.exp(x-y)); //= y + log(1+e^{x-y}) = log(e^{y}) + log(1+e{x-y}) = log(e^{y} * (1 + e^{x-y})) = log(e^{y} + e^{x})
		else 
			return x + Math.log1p(Math.exp(y-x)); //Same as above, but reversed
	}

	public int sampleX(double[] potentialArray) {
//		return sampleEntry(potentialArray);
		return sampleEntry2(potentialArray);
	}

	
	/*
	 * Sample an entry based on scores.
	 * 
	 * All the scores should be >=0. 0 means it is impossible. The sum of all entries should be provided in sumScores
	 * 
	 */
	public static int rouletteSampleEntry(Random random,double[] scores,double sumScores){
		double d=random.nextDouble()*sumScores;
		
		double sum=0.d;
		
		for(int i=0;i<scores.length;++i){
			sum+=scores[i];
			if(d<sum)
				return i;
		}
		
		return scores.length-1; //This is should only happen with floating point issues where we are very close to the sum
	}
}
