package ims.util;

import java.util.regex.Pattern;

public class StringShape {

	static final String UPPERCASE_REPLACE="A";
	static final String LOWERCASE_REPLACE="a";
	static final String DIGIT_REPLACE="0";
	static final String OTHER_REPLACE=".";
	static final String MORE_THAN_ONE_REPLACE=" "; //Use space here. We assume there are no spaces in our inputs.
	
	static final Pattern ONE_UPPERCASE=Pattern.compile("[A-Z]");
	static final Pattern ONE_LOWERCASE=Pattern.compile("[a-z]");
	static final Pattern ONE_DIGIT=Pattern.compile("[0-9]");
	static final Pattern OTHER_AFTER_INIT_REPLACE=Pattern.compile("[^Aa0 ]");
			
	static final Pattern MORE_THAN_ONE_UPPERCASE=Pattern.compile("[A-Z]{2,}");
	static final Pattern MORE_THAN_ONE_LOWERCASE=Pattern.compile("[a-z]{2,}");
	static final Pattern MORE_THAN_ONE_DIGIT=Pattern.compile("[0-9]{2,}");
	static final Pattern MORE_THAN_ONE_OTHER_AFTER_INIT_REPLACE=Pattern.compile("[^Aa0 ]{2,}");
	
	static final String MORE_THAN_ONE_UPPERCASE_REPLACE=UPPERCASE_REPLACE+MORE_THAN_ONE_REPLACE;
	static final String MORE_THAN_ONE_LOWERCASE_REPLACE=LOWERCASE_REPLACE+MORE_THAN_ONE_REPLACE;
	static final String MORE_THAN_ONE_DIGIT_REPLACE=DIGIT_REPLACE+MORE_THAN_ONE_REPLACE;
	static final String MORE_THAN_ONE_OTHER_REPLACE=OTHER_REPLACE+MORE_THAN_ONE_REPLACE;

	public static String reduceToShape(String s){
		//More than one
		s=MORE_THAN_ONE_UPPERCASE.matcher(s).replaceAll(MORE_THAN_ONE_UPPERCASE_REPLACE);
		s=MORE_THAN_ONE_LOWERCASE.matcher(s).replaceAll(MORE_THAN_ONE_LOWERCASE_REPLACE);
		s=MORE_THAN_ONE_DIGIT.matcher(s).replaceAll(MORE_THAN_ONE_DIGIT_REPLACE);
		
		//Just one
		s=ONE_UPPERCASE.matcher(s).replaceAll(UPPERCASE_REPLACE);
		s=ONE_LOWERCASE.matcher(s).replaceAll(LOWERCASE_REPLACE);
		s=ONE_DIGIT.matcher(s).replaceAll(DIGIT_REPLACE);

		//And everything non-alphanumeric
		s=MORE_THAN_ONE_OTHER_AFTER_INIT_REPLACE.matcher(s).replaceAll(MORE_THAN_ONE_OTHER_REPLACE);
		s=OTHER_AFTER_INIT_REPLACE.matcher(s).replaceAll(OTHER_REPLACE);
		
		return s;
	}
	
	public static String getCharShape(char c){
		String s=new String(new char[]{c});
		if(ONE_UPPERCASE.matcher(s).matches())
			return UPPERCASE_REPLACE;
		else if(ONE_LOWERCASE.matcher(s).matches())
			return LOWERCASE_REPLACE;
		else if(ONE_DIGIT.matcher(s).matches())
			return DIGIT_REPLACE;
		else 
			return OTHER_REPLACE;
	}
	
	public static int getCharShapeInt(char c){
		switch(getCharShape(c)){
		case UPPERCASE_REPLACE: return 2;
		case LOWERCASE_REPLACE: return 3;
		case DIGIT_REPLACE:     return 4;
		case OTHER_REPLACE:     return 5;
		default:
			throw new Error("not defined");
		}
	}
	
	public static void main(String[] args){
		String[] toTest={
				"Anders","ANders","ANDERS","0Anders","35235An2424,.++DErsAA",
				"<<","File",":","Questar","So","Trails","Hub","Interconnect","02-05-02.doc",">>",
		};
		
		for(String s:toTest){
			String shape=reduceToShape(s);
			System.out.printf("%-30s --> %-10s\n","'"+s+"'","'"+shape+"'");
		}
		
		int times=1000*1000;
		
		System.out.println();
		System.out.println();
		System.out.println("Now benchmark: "+times+" times to do the shape.");
		
		long start=System.currentTimeMillis();
		for(int i=0;i<times;++i){
			for(String s:toTest){
				reduceToShape(s);
			}
		}
		long time=System.currentTimeMillis()-start;
		System.out.println();
		System.out.println("Time needed: "+Util.insertCommas(time, true));
	}
}
