package ims.util.visualtrans;

import ims.data.CoNLL06Sentence;
import ims.data.ISentence;
import ims.oracle.Configuration;
import ims.oracle.PlotOracleLattice;
import ims.oracle.Configuration.TR;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.util.visualtrans.ParseContext.ColorCoding;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

public class VisualizationServlet extends HttpServlet {
	private static final long serialVersionUID = 3945758096792886783L;

	static {
		PlotOracleLattice.VERBOSE_ERRORS=false;
	}
	
	public static enum Highlight {
		None,Last,NonProj,LastAndNonProj;
	}
	
	
	static final String HTML_HEADER="<html><head><title>ITSViz</title></head><body>\n"
//			+ "<meta name=\"accept-encoding\" value=\"utf8\"/>"
			+ "<style type=\"text/css\">\n"
			+ "a:link, a:visited { color: #0000FF; text-decoration: none; }\n"
			+ ".graybox {background: #CCCCCC; }\n"
			+ ".scrollit {overflow:auto; width:95%;}\n"
			+ ".stackTop  { text-align: center; border-right: 1px solid black; border-left: 1px solid black; border-top: 1px solid black; }\n"
			+ ".stackItem { text-align: center; border-right: 1px solid black; border-left: 1px solid black; }\n"
			+ ".buffer {border-left:1px solid black; border-top: 1px solid black; border-bottom: 1px solid black; }\n"
			+ ".SH { background-color: "+TR.SH.htmlColor+"; }\n"
			+ ".LA { background-color: "+TR.LA.htmlColor+"; }\n"
			+ ".RAS { background-color: "+TR.RAS.htmlColor+"; }\n"
			+ ".SW { background-color: "+TR.SW.htmlColor+"; }\n"
			+ ""
			+ ""
			+ "</style>\n";
	
	static final String HTML_FOOTER="</body></html>\n";
	
	static final String INPUT_TEXT="InputText";
	static final String INPUT_ACTION ="action";
	static final String INPUT_BACTION = "baction";
	static final String INPUT_PNG_STATE_SIG="pngStateSig";
	static final String INPUT_TRANS = "tr";
	static final String INPUT_LAST_SIG="lastSig";
	static final String INPUT_STATE_SIG="sig";
	static final String INPUT_SYSTEM_SELECT = "systemSelect";
	static final String INPUT_SUBSTATE="substate";
	static final String INPUT_CURVED="curved";
	static final String INPUT_DISPLAY_STACK_TOKENS="stackTokens";
	static final String INPUT_COLOR_CODING="colorCoding";
	
	static final String ACTION_INIT="init";
	static final String ACTION_APPLY_TR="applyTR";
	static final String ACTION_GET_PNG="getPNG";
	static final String ACTION_DISPLAY_STATE="displayState";
	static final String ACTION_REFERENCE_PARSE="referenceParse";
	
	static final String BACTION_UPDATE_CONFIG = "updateConfig";
	
	static final String SESSION_PARSE_CONTEXT_KEY="parseContext";
 
	static final Map<Character,TR> k2TRLookup;
	static final Map<String,TR[]> system2TRLookup;

	

	
	
	static {
		k2TRLookup=new HashMap<Character,TR>();
		for(TR tr:TR.values())
			k2TRLookup.put(tr.k, tr);
		
		system2TRLookup=new TreeMap<String,TR[]>();
		system2TRLookup.put("ArcStandard",      TransitionSystem.ArcS.trs);
		system2TRLookup.put("ArcStandard+Swap", TransitionSystem.ArcSNewSwap.trs);
		system2TRLookup.put("All (debug)",      TR.values());
	}
	
	
	public void init(){

	}
	
	private void handle(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException{
		request.setCharacterEncoding("UTF-8"); //Should work if called before anything gets read (according to http://stackoverflow.com/questions/8391675/utf-8-encoding-a-servlet-form-submission-with-tomcat
//		response.setHeader("Accept-encoding", "charset=utf8");
		System.out.println(request.getCharacterEncoding());
		System.out.println(request.getContentType());
		String action=request.getParameter(INPUT_ACTION);
		String text=request.getParameter(INPUT_TEXT);
		String sys=request.getParameter(INPUT_SYSTEM_SELECT);
		boolean stackTokens=Boolean.parseBoolean(request.getParameter(INPUT_DISPLAY_STACK_TOKENS));
		boolean curved=Boolean.parseBoolean(request.getParameter(INPUT_CURVED));
		
		String subState=request.getParameter(INPUT_SUBSTATE);
		int ss=(subState!=null?Integer.parseInt(subState):-1);
		
		ParseContext pc=(ParseContext) request.getSession().getAttribute(SESSION_PARSE_CONTEXT_KEY);
		
		String bAction=request.getParameter(INPUT_BACTION);
		if(bAction!=null){
			if(bAction.equals(BACTION_UPDATE_CONFIG)){
				pc.trs=system2TRLookup.get(sys);
				pc.setCurved(curved);
			}
		}
		
		if(action==null){
			printInitPage(response);
		} else if(action.equals(ACTION_INIT)){
			pc=initParseContext(text,system2TRLookup.get(sys),curved);
			request.getSession().setAttribute(SESSION_PARSE_CONTEXT_KEY, pc);
			respond(response,pc,-1,stackTokens);
		} else if(action.equals(ACTION_APPLY_TR)){
			String tr=request.getParameter(INPUT_TRANS);
			String lastSig=request.getParameter(INPUT_LAST_SIG);
			if(ss>=0 && lastSig.length()>ss+2 && lastSig.substring(0,ss+1+1).endsWith(Character.toString(TR.valueOf(tr).k))){
				ss++;
			} else {
				String applySig=lastSig;
				if(ss>=0)
					applySig=lastSig.substring(0, ss+1);
				pc.applyTr(TR.valueOf(tr),applySig);
				ss=-1;
			}
			respond(response,pc,ss,stackTokens);
		} else if(action.equals(ACTION_REFERENCE_PARSE)){
			pc=oracleParse(text,sys,curved);
			request.getSession().setAttribute(SESSION_PARSE_CONTEXT_KEY, pc);
			respond(response,pc,-1,stackTokens);
		} else if(action.equals(ACTION_GET_PNG)){
			servePNG(request,response);
		} else if(action.equals(ACTION_DISPLAY_STATE)){
			String sig=request.getParameter(INPUT_STATE_SIG);
			pc.setCurrent(sig);
			respond(response,pc,ss,stackTokens);
		} else {
			printInitPage(response);
		}

	}

	public void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException{
		handle(request,response);
	}

	public void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException{
		handle(request,response);
	}

	private void printInitPage(HttpServletResponse response) throws IOException {
		response.setContentType("text/html;charset=utf-8");
		PrintWriter out = response.getWriter();
		out.println(HTML_HEADER);
		out.println("<h1>Play</h1>");
		out.println("<form action=\".\" method=\"post\">");
		out.println("<input type=\"hidden\" name=\""+INPUT_ACTION+"\" value=\""+ACTION_INIT+"\"/>");
		out.println("<textarea name=\""+INPUT_TEXT+"\" rows=10 cols=80></textarea><br/>");
		out.println("<select name=\""+INPUT_SYSTEM_SELECT+"\">");
		for(Entry<String,TR[]> e:system2TRLookup.entrySet())
			out.println("<option value=\""+e.getKey()+"\">"+e.getKey()+"</option>");
		out.println("</select>");
		out.println("<select name=\""+INPUT_CURVED+"\"><option value=\"false\">Square</option><option value=\"true\">Curved</option></select>");
		out.println("<input type=\"submit\"/>");
		out.println("</form>\n");
		
		out.println("<hr/>");
		out.println("<h1>Oracle parse</h1>");
		out.println("<form action=\".\" method=\"post\">");
		out.println("<input type=\"hidden\" name=\""+INPUT_ACTION+"\" value=\""+ACTION_REFERENCE_PARSE+"\"/>");
		out.println("<textarea name=\""+INPUT_TEXT+"\" rows=20 cols=160></textarea><br/>");
		out.println("<select name=\""+INPUT_SYSTEM_SELECT+"\">");
		for(TransitionSystem ts:Options.TransitionSystem.values())
			out.println("<option value=\""+ts.toString()+"\">"+ts.toString()+"</option>");
		out.println("</select>");
		out.println("<select name=\""+INPUT_CURVED+"\"><option value=\"false\">Square</option><option value=\"true\">Curved</option></select>");
		out.println("<input type=\"submit\"/>");
		out.println("</form>\n");

		//Now the stuff for oracle parsing
		out.println(HTML_FOOTER);
	}
	
	private void servePNG(HttpServletRequest request,HttpServletResponse response) throws IOException {
		HttpSession session=request.getSession();
		String stateSig=request.getParameter(INPUT_PNG_STATE_SIG);
		boolean curved=Boolean.parseBoolean(request.getParameter(INPUT_CURVED));
		ColorCoding cc=ColorCoding.valueOf(request.getParameter(INPUT_COLOR_CODING));
		response.setContentType("image/png");
		ParseContext pc=(ParseContext) session.getAttribute(SESSION_PARSE_CONTEXT_KEY);
		byte[] b=pc.getPNG(stateSig,curved,cc);
		response.getOutputStream().write(b);
	}

	private static ISentence getSentence(String text) {
		List<String> lines=Arrays.asList(text.split("[\r]?\n"));
		return CoNLL06Sentence.lines2Sentence(lines, 0);
	}

	private void respond(HttpServletResponse response, ParseContext pc,int subState,boolean displayStateTokenForms) throws IOException {
		response.setContentType("text/html;charset=utf-8");
//		String sig=pc.current.getSignature();
		String stateSig=pc.current.getTransitionSeqSignature();
		if(subState>=0)
			stateSig=stateSig.substring(0, subState+1);
		PrintWriter out = response.getWriter();
		out.println(HTML_HEADER);
		Configuration displayConfig=pc.current;
		if(subState>=0 && subState<pc.current.getTransitionSeqSignature().length()-1)
			displayConfig=pc.configs.get(pc.current.getTransitionSeqSignature().substring(0, 1+subState));
		
		//The picture
		out.println("<h4>Tree</h4>");
		out.println("<div class=\"scrollit\">");
		String pngUrl="?"+INPUT_ACTION+"="+ACTION_GET_PNG+"&"+INPUT_PNG_STATE_SIG+"="+stateSig+"&"+INPUT_CURVED+"="+pc.curved+"&"+INPUT_COLOR_CODING+"="+pc.cc.toString();
		out.println("<img src=\""+pngUrl+"\" />");
		out.println("</div>");

		//Buttons to apply an action
		out.println("<form action=\".\" method=\"post\">");
		out.println("<input type=\"hidden\" name=\""+INPUT_ACTION+"\" value=\""+ACTION_APPLY_TR+"\"/>");
		out.println("<input type=\"hidden\" name=\""+INPUT_LAST_SIG+"\" value=\""+pc.current.getTransitionSeqSignature()+"\"/>");
		out.println("<input type=\"hidden\" name=\""+INPUT_SUBSTATE+"\" value=\""+subState+"\"/>");
		out.println("<input type=\"hidden\" name=\""+INPUT_DISPLAY_STACK_TOKENS+"\" value=\""+displayStateTokenForms+"\"/>");
		out.println("<table cellpadding=10 cellspacing=2>");
		out.println("<tr>");
		for(TR tr:pc.trs){
			String disabled="disabled=\"true\"";
			if(displayConfig.isPermissible(tr))
				disabled="";
			out.println("<td class=\""+tr.toString()+"\"><button type=\"submit\" name=\""+INPUT_TRANS+"\" value=\""+tr.toString()+"\" "+disabled+" >"+tr.name+"</button></td>");
		}
		out.println("</tr>");
		out.println("</table>");
		out.println("</form>");

		
		//Display history
		String sig=pc.current.getTransitionSeqSignature();
		
		String prevStateURL=null;
		String nextStateURL=null;
		String fastForwardURL=null;
		String fastBackwardURL=null;
		out.println("<br/>");
		out.println("<h4>History</h4>");
		
		int totalState=sig.length();
		int highlightState=(subState==-1?sig.length():subState);
		out.printf("<div class=\"scrollit\" id=\"historydiv\">", highlightState,totalState);
		out.println();
		out.println("<table cellpadding=8 border=1 cellspacing=0><tr>");
		for(int i=0;i<sig.length();++i){
			String url="?action="+ACTION_DISPLAY_STATE+"&"+INPUT_DISPLAY_STACK_TOKENS+"="+displayStateTokenForms+"&"+INPUT_STATE_SIG+"="+sig+"&"+INPUT_SUBSTATE+"="+i;
			String tdDecoration="";
			if(i==0)
				fastBackwardURL=url;
			if(subState==-1){
				//Highlight current
				if(i==sig.length()-1)
					tdDecoration=" class=\"graybox\"";
				//Track the previous
				if(i==sig.length()-2)
					prevStateURL=url;
			} else {
				//Highlight current
				if(i==subState)
					tdDecoration=" class=\"graybox\"";
				//Track previous
				if(i==subState-1)
					prevStateURL=url;
				//Track next
				if(i==subState+1)
					nextStateURL=url;
				if(i==sig.length()-1)
					fastForwardURL=url;
			}
			if(i==subState-1)
				prevStateURL=url;
			if(i==subState+1 && subState>=0)
				nextStateURL=url;
			out.print("<td"+tdDecoration+"><a href=\""+url+"\">c<sub>"+i+"</sub></a></td>");
			
		}
		out.println("</tr><tr>");
		for(int i=1;i<sig.length();++i){
			String clazz;
			if((subState>=0 && i==subState || subState==-1 && i==sig.length()-1)){
				clazz="graybox";
			} else {
				clazz=k2TRLookup.get(Character.toString(sig.charAt(i))).toString();
			}
			out.print("<td class=\""+clazz+"\">"+k2TRLookup.get(sig.substring(i, i+1)).toString()+"</td>");
		}
		out.println("<td>&nbsp;</td></tr></table></div>");
		//Focus the scrollable div
		out.println("<script type=\"text/javascript\">");
		out.println("  d=document.getElementById('historydiv');");
		out.printf("  d.scrollLeft=%d/%d*d.scrollWidth-0.5*d.offsetWidth;\n", (highlightState+1),totalState);
//		 onLoad='scrollLeft=%d/%d*scrollWidth-0.5*offsetWidth'
		out.println("</script>");
		out.println("<center>");
		out.println("<table cellpadding=10><tr>");
		if(fastBackwardURL!=null)
			out.println("<td><form action=\"."+fastBackwardURL+"\" method=\"post\"><input type=\"submit\" value=\"Back to Start\"></form></td>");
		if(prevStateURL!=null)
			out.println("<td><form action=\"."+prevStateURL+"\" method=\"post\"><input type=\"submit\" value=\"Back\"></form></td>");
		if(nextStateURL!=null)
			out.println("<td><form action=\"."+nextStateURL+"\" method=\"post\"><input type=\"submit\" value=\"Forward\"></form></td>");
		if(fastForwardURL!=null)
			out.println("<td><form action=\"."+fastForwardURL+"\" method=\"post\"><input type=\"submit\" value=\"Forward to Last\"></form></td>");
		out.println("</tr></table>");
		out.println("</center>");
		
		//Display current state
		out.println("<br/>");
		out.println("<h4>State</h4>");
		out.println("<table cellpadding=4 cellspacing=0>");
		out.println("<tr><th>&nbsp;</th><th>&nbsp;</th><th>Buffer</th></tr>");
		String buffer;
		if(displayStateTokenForms)
			buffer=displayConfig.bufferToString(pc.t);
		else
			buffer=displayConfig.bufferToString();
		out.println("<tr><td>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td class=\"buffer\"><tt>"+buffer+"</tt></td></tr>");
		for(int i=0;i<displayConfig.stackSize();++i){
			String item=(displayStateTokenForms?pc.t[displayConfig.peekStack(i)].replaceAll("<", "&lt;").replaceAll(">", "&gt"):Integer.toString(displayConfig.peekStack(i)));
			out.println("<tr><td class=\""+(i==0?"stackTop":"stackItem")+"\"><tt>"+item+"</tt></td><td>&nbsp;</td><td>&nbsp;</td></tr>");
		}
		out.println("<tr><th>Stack</th><th>&nbsp</th></tr>");
		out.println("</table>");
		
		//Options
		out.println("<br>");
		out.println("<h4>Options</h4>");
		out.println("<form action=\".\" method=\"post\">");
		out.println("<input type=hidden name=\"action\" value=\""+ACTION_DISPLAY_STATE+"\"/>");
		out.println("<input type=hidden name=\"baction\" value=\""+BACTION_UPDATE_CONFIG+"\"/>");
		out.println("<input type=hidden name=\""+INPUT_STATE_SIG+"\" value=\""+pc.current.getTransitionSeqSignature()+"\"/>");
		out.println("<input type=hidden name=\""+INPUT_SUBSTATE+"\" value=\""+subState+"\"/>");
		out.println("<table cellpadding=8>");
		out.println("<tr><td><select name=\""+INPUT_CURVED+"\"><option value=\"true\">Curved</option><option value=\"false\">Square</option></select></td></tr>");
		out.println("<tr><td><select name=\""+INPUT_DISPLAY_STACK_TOKENS+"\"><option value=\"true\">Forms</option><option value=\"false\">IDs</option></select></td></tr>");
		out.println("<tr><td>");
		out.println("<select name=\""+INPUT_SYSTEM_SELECT+"\">");
		for(Entry<String,TR[]> e:system2TRLookup.entrySet())
			out.println("<option value=\""+e.getKey()+"\">"+e.getKey()+"</option>");
		out.println("</select></td></tr>");
		out.println("<tr><td><input type=\"submit\"/></td></tr></table>");
		out.println("</form>");

		//Back to start
		out.println("<br/><br/><a href=\".\">Back to start</a>");
		out.println(HTML_FOOTER);
	}

	private ParseContext initParseContext(String text,TR[] trs,boolean curved) throws IOException {
		String[] tokens=text.split("\\s+");
		List<String> t=new ArrayList<String>(Arrays.asList(tokens));
		ParseContext pc=new ParseContext(t,trs,curved,ColorCoding.LastAndNonProj);
		return pc;
	}

	
	private ParseContext oracleParse(String text, String ts,boolean curved) throws IOException {
		text=text.replaceAll("\\s+$", "");
		ISentence s=getSentence(text);
		List<String> tokens=new ArrayList<String>();
		String[] form=s.getForms();
		
		for(int i=1;i<form.length;++i)
			tokens.add(form[i]);
		IOracle oracle=OracleFactory.getStaticOracle(TransitionSystem.valueOf(ts));
		ParseContext pc=new ParseContext(tokens,TransitionSystem.valueOf(ts).trs,curved,ColorCoding.LastAndNonProj);
		List<String> ops=oracle.getCanonicalSequence(s.getGoldHeads(), s.getGoldLabels(), false);
		for(String op:ops){
			TR tr=TR.valueOf(op);
			pc.applyTr(tr, pc.current.getTransitionSeqSignature());
		}
		return pc;
	}
}
